# Copyright (c) 2024-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import os

import astropy.units as u
import numpy as np

import toast
from toast.utils import Logger
from toast.traits import trait_docs, Int, Unicode, Instance, List, Bool
from toast.timing import function_timer
from toast.intervals import IntervalList
from toast.observation import default_values as defaults
from toast.ops import Operator
from toast import qarray as qa


class Split(object):
    """Base class for objects implementing a particular split.

    Derived classes should implement the `_create_split()` method and optionally
    the `_remove_split()` method.

    When the `_create_split` method is called, the detector flags have already
    been backed up and the derived class can modify them to exclude detectors by
    setting the invalid mask bit.  If the derived class does not set the split
    sample intervals, then all samples will be used.  Many splits simply make a
    copy of an existing interval list to use for the split.  However, anything
    could be done here, including logical operations with multiple interval lists
    or constructing a new interval list from other metadata.

    The `_remove_split` method is only needed if any other observation cleanup
    is needed.  After this method is called, the splits interval will be deleted
    and the detector flags restored.

    """

    def __init__(self, name="N/A"):
        self._name = name
        self._split_intervals = "split"
        self._saved_det_flags = "saved_det_split_flags"

    @property
    def name(self):
        return self._name

    @property
    def split_intervals(self):
        return self._split_intervals

    def _create_split(self, obs):
        pass

    def _remove_split(self, obs):
        pass

    def _delete_split(self, obs):
        """Remove any split products that exist."""
        if self._split_intervals in obs.intervals:
            del obs.intervals[self._split_intervals]
        if self._saved_det_flags in obs:
            del obs[self._saved_det_flags]

    def create_split(self, obs):
        """Create the split in the given observation.

        Args:
            obs (Observation):  The observation to modify.

        Returns:
            None

        """
        # Clear any existing split products.  These should not exist
        # if previous splits were properly removed.
        self._delete_split(obs)

        # Backup detector flags
        obs[self._saved_det_flags] = dict(obs.local_detector_flags)

        # Call split-specific creation
        self._create_split(obs)

        # If the split interval was not setup by the derived class, set it
        # to None (all samples).
        if self._split_intervals not in obs.intervals:
            obs.intervals[self._split_intervals] = IntervalList(
                intervals=obs.intervals[None]
            )

    def remove_split(self, obs):
        """Remove the split from the given observation.

        Args:
            obs (Observation):  The observation to modify.

        Returns:
            None

        """
        # Perform any split-specific operations
        self._remove_split(obs)

        # Restore detector flags and remove the splits interval.
        obs.set_local_detector_flags(obs[self._saved_det_flags])
        self._delete_split(obs)

    @staticmethod
    def create(name, **kwargs):
        """Factory method to instantiate splits.

        The name is the hard-coded name of each split derived class.
        any kwargs are passed to the constructor.  This function is
        the one place to register new split types and their names.

        Args:
            name (str):  The name of the split.  If a list is passed, then
                the SplitByList split is created.

        """
        log = Logger.get()
        if isinstance(name, list):
            # The "real" name should be in kwargs
            split_name = kwargs["split_name"]
            del kwargs["split_name"]
            kwargs["dets"] = name
            return SplitByList(split_name, **kwargs)
        elif name == "all":
            return SplitAll(name, **kwargs)
        elif name == "left_going":
            return SplitLeftGoing(name, **kwargs)
        elif name == "right_going":
            return SplitRightGoing(name, **kwargs)
        elif name == "outer_detectors":
            return SplitOuterDetectors(name, **kwargs)
        elif name == "inner_detectors":
            return SplitInnerDetectors(name, **kwargs)
        elif name == "polA":
            return SplitPolA(name, **kwargs)
        elif name == "polB":
            return SplitPolB(name, **kwargs)
        else:
            msg = f"Unsupported split '{name}'"
            log.error(msg)
            raise RuntimeError(msg)


class SplitAll(Split):
    """Split to process all data.

    Args:
        name (str):  Name of the split.
        interval (str):  Name of the interval to use.

    """

    def __init__(self, name, interval=defaults.scanning_interval):
        super().__init__(name)
        self._interval = interval

    def _create_split(self, obs):
        timespans = [(x.start, x.stop) for x in obs.intervals[self._interval]]
        obs.intervals[self._split_intervals] = IntervalList(
            obs.shared[defaults.times],
            timespans=timespans,
        )


class SplitLeftGoing(Split):
    """Split to process only left-going scans.

    Args:
        name (str):  Name of the split.
        interval (str):  Name of the interval to use.

    """

    def __init__(self, name, interval=defaults.scan_rightleft_interval):
        super().__init__(name)
        self._interval = interval

    def _create_split(self, obs):
        timespans = [(x.start, x.stop) for x in obs.intervals[self._interval]]
        obs.intervals[self._split_intervals] = IntervalList(
            obs.shared[defaults.times],
            timespans=timespans,
        )


class SplitRightGoing(Split):
    """Split to process only right-going scans.

    Args:
        name (str):  Name of the split.
        interval (str):  Name of the interval to use.

    """

    def __init__(self, name, interval=defaults.scan_leftright_interval):
        super().__init__(name)
        self._interval = interval

    def _create_split(self, obs):
        timespans = [(x.start, x.stop) for x in obs.intervals[self._interval]]
        obs.intervals[self._split_intervals] = IntervalList(
            obs.shared[defaults.times],
            timespans=timespans,
        )


class SplitOuterDetectors(Split):
    """Split to process only detectors on the edge of the focalplane.

    Args:
        name (str):  Name of the split.
        radial_cut (float):  Radial distance from boresight to cut in
            units of FOV/2.  Detectors inside this are cut.
        interval (str):  Name of the interval to use.

    """

    def __init__(
        self,
        name,
        radial_cut=0.7071067811865476,
        interval=defaults.scanning_interval,
    ):
        super().__init__(name)
        self._radial_cut = radial_cut
        self._interval = interval

    def _create_split(self, obs):
        focalplane = obs.telescope.focalplane
        fp_radius = focalplane.field_of_view / 2
        limit = fp_radius.to_value(u.rad) * self._radial_cut

        # Flag detectors
        flags = dict(obs.local_detector_flags)
        for det in obs.select_local_detectors():
            det_quat = focalplane[det]["quat"]
            det_theta, det_phi, det_psi = qa.to_iso_angles(det_quat)
            if det_theta < limit:
                # Cut the detector
                flags[det] |= defaults.det_mask_invalid
        obs.set_local_detector_flags(flags)

        # Set the split intervals to be a simple copy of the input interval list.
        timespans = [(x.start, x.stop) for x in obs.intervals[self._interval]]
        obs.intervals[self._split_intervals] = IntervalList(
            obs.shared[defaults.times],
            timespans=timespans,
        )


class SplitInnerDetectors(Split):
    """Split to process only detectors on the inner part of the focalplane.

    Args:
        name (str):  Name of the split.
        radial_cut (float):  Radial distance from boresight to cut in
            units of FOV/2.  Detectors outside this are cut.
        interval (str):  Name of the interval to use.

    """

    def __init__(
        self,
        name,
        radial_cut=0.7071067811865476,
        interval=defaults.scanning_interval,
    ):
        super().__init__(name)
        self._radial_cut = radial_cut
        self._interval = interval

    def _create_split(self, obs):
        focalplane = obs.telescope.focalplane
        fp_radius = focalplane.field_of_view / 2
        limit = fp_radius.to_value(u.rad) * self._radial_cut

        # Flag detectors
        flags = dict(obs.local_detector_flags)
        for det in obs.select_local_detectors():
            det_quat = focalplane[det]["quat"]
            det_theta, det_phi, det_psi = qa.to_iso_angles(det_quat)
            if det_theta >= limit:
                # Cut the detector
                flags[det] |= defaults.det_mask_invalid
        obs.set_local_detector_flags(flags)

        # Set the split intervals to be a simple copy of the input interval list.
        timespans = [(x.start, x.stop) for x in obs.intervals[self._interval]]
        obs.intervals[self._split_intervals] = IntervalList(
            obs.shared[defaults.times],
            timespans=timespans,
        )


class SplitByList(Split):
    """Split to process only an explicit list of detectors.

    If dets is None, all detectors are used.

    Args:
        name (str):  Name of the split.
        dets (list) : List of detectors to process.
        interval (str):  Name of the interval to use.

    """

    def __init__(
        self,
        name,
        dets=None,
        interval=defaults.scanning_interval,
    ):
        super().__init__(name)
        # Convert to a set for faster lookup.
        if dets is None:
            self._dets = None
        else:
            self._dets = set(dets)
        self._interval = interval
        # Add all possible demodulated names of the detector
        demod_dets = set()
        for det in self._dets:
            for prefix in ["demod0", "demo2r", "demod2i", "demod4r", "demod4i"]:
                demod_dets.add(f"{prefix}_{det}")
        self._dets.update(demod_dets)

    def _create_split(self, obs):
        # Flag detectors
        if self._dets is not None:
            # We have some selection
            flags = dict(obs.local_detector_flags)
            for det in obs.select_local_detectors():
                if det not in self._dets:
                    # Not in the list, cut it.
                    flags[det] |= defaults.det_mask_invalid
            obs.set_local_detector_flags(flags)

        # Set the split intervals to be a simple copy of the input interval list.
        timespans = [(x.start, x.stop) for x in obs.intervals[self._interval]]
        obs.intervals[self._split_intervals] = IntervalList(
            obs.shared[defaults.times],
            timespans=timespans,
        )


class SplitPolA(Split):
    """Split to process only detector with A polarization.

    Args:
        name (str):  Name of the split.
        interval (str):  Name of the interval to use.

    """

    def __init__(self, name, interval=defaults.scanning_interval):
        super().__init__(name)
        self._interval = interval

    def _create_split(self, obs):
        focalplane = obs.telescope.focalplane

        # Flag detectors
        flags = dict(obs.local_detector_flags)
        for det in obs.select_local_detectors():
            pol = focalplane[det]["pol"]
            if pol != "A":
                # Cut the detector
                flags[det] |= defaults.det_mask_invalid
        obs.set_local_detector_flags(flags)

        # Set the split intervals to be a simple copy of the input interval list.
        timespans = [(x.start, x.stop) for x in obs.intervals[self._interval]]
        obs.intervals[self._split_intervals] = IntervalList(
            obs.shared[defaults.times],
            timespans=timespans,
        )


class SplitPolB(Split):
    """Split to process only detector with B polarization.

    Args:
        name (str):  Name of the split.
        interval (str):  Name of the interval to use.

    """

    def __init__(self, name, interval=defaults.scanning_interval):
        super().__init__(name)
        self._interval = interval

    def _create_split(self, obs):
        focalplane = obs.telescope.focalplane

        # Flag detectors
        flags = dict(obs.local_detector_flags)
        for det in obs.select_local_detectors():
            pol = focalplane[det]["pol"]
            if pol != "B":
                # Cut the detector
                flags[det] |= defaults.det_mask_invalid
        obs.set_local_detector_flags(flags)

        # Set the split intervals to be a simple copy of the input interval list.
        timespans = [(x.start, x.stop) for x in obs.intervals[self._interval]]
        obs.intervals[self._split_intervals] = IntervalList(
            obs.shared[defaults.times],
            timespans=timespans,
        )


@trait_docs
class Splits(Operator):
    """Apply a sequence of splits to the data, and make a map of each.

    The list of `splits` contains strings, where each string is either
    the name of the split, or is the name of the split followed by a
    string representation of a dictionary to pass as kwargs to the split
    constructor.

    """

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    mapmaker = Instance(
        klass=Operator,
        allow_none=True,
        help="The mapmaking operator to use",
    )

    output_dir = Unicode(
        ".",
        help="Top-level output directory",
    )

    splits = List([], help="The list of named splits to apply")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        return

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()

        if self.mapmaker is None:
            msg = "You must specify the mapmaker operator"
            log.error(msg)
            raise RuntimeError(msg)

        # Build splits we are using
        self._split_obj = dict()
        for split_arg in self.splits:
            # Each element of the list might be just a name, or it might
            # be a name followed by a dictionary of kwargs.
            split_opts = split_arg.split(":")
            if len(split_opts) > 1:
                split_name = split_opts[0]
                # Join the rest of the string, use python eval to convert
                # to a dictionary.
                split_kw = dict()
                raise NotImplementedError("Parsing of extra split kwargs not complete")
            else:
                split_name = split_opts[0]
                split_kw = dict()
            if os.path.isfile(split_name):
                # This is a file containing the list of detectors to use.
                # Build the split name from the path.
                name, ext = os.path.splitext(os.path.basename(split_name))
                det_list = None
                if data.comm.world_rank == 0:
                    det_list = list()
                    with open(split_name, "r") as f:
                        for line in f:
                            det_list.append(line.strip())
                if data.comm.comm_world is not None:
                    det_list = data.comm.comm_world.bcast(det_list, root=0)
                split_kw["split_name"] = name
                self._split_obj[name] = Split.create(det_list, **split_kw)
            else:
                # Just a normal split type
                self._split_obj[split_name] = Split.create(split_name, **split_kw)

        if data.comm.world_rank == 0:
            msg = "Using splits: "
            for sname in self._split_obj.keys():
                msg += f"{sname}, "
            log.info(msg)

        # Save starting mapmaker parameters, to restore later
        mapmaker_save_traits = dict()
        for trait_name, trait in self.mapmaker.traits().items():
            mapmaker_save_traits[trait_name] = trait.get(self.mapmaker)

        # Possible traits that we want to toggle off
        mapmaker_disable_traits = [
            "write_binmap",
            "write_map",
            "write_noiseweighted_map",
            "write_hits",
            "write_cov",
            "write_invcov",
            "write_rcond",
            "write_solver_products",
            "save_cleaned",
        ]

        # Possible traits we want to enable
        mapmaker_enable_traits = [
            "keep_final_products",
            "reset_pix_dist",
        ]

        if hasattr(self.mapmaker, "map_binning"):
            map_binner = self.mapmaker.map_binning
        else:
            map_binner = self.mapmaker.binning
        pointing_view_save = map_binner.pixel_pointing.view

        # If the pixel distribution does yet exit, we create it once prior to
        # doing any splits.
        if map_binner.pixel_dist not in data:
            pix_dist = toast.ops.BuildPixelDistribution(
                pixel_dist=map_binner.pixel_dist,
                pixel_pointing=map_binner.pixel_pointing,
                save_pointing=map_binner.full_pointing,
            )
            pix_dist.apply(data)

        for trt in mapmaker_disable_traits:
            if hasattr(self.mapmaker, trt):
                setattr(self.mapmaker, trt, False)

        for trt in mapmaker_enable_traits:
            if hasattr(self.mapmaker, trt):
                setattr(self.mapmaker, trt, True)

        # Loop over splits
        for split_name, spl in self._split_obj.items():
            log.info_rank(f"Running Split '{split_name}'", comm=data.comm.comm_world)
            # Set mapmaker name based on split and the name of this
            # Splits operator.
            mname = f"{self.name}_{split_name}"
            self.mapmaker.name = mname

            # Apply this split
            for ob in data.obs:
                spl.create_split(ob)

            # Set mapmaking tools to use the current split interval list
            map_binner.pixel_pointing.view = spl.split_intervals
            if not map_binner.full_pointing:
                # We are not using full pointing and so we clear the
                # residual pointing for this split
                toast.ops.Delete(
                    detdata=[
                        map_binner.pixel_pointing.pixels,
                        map_binner.stokes_weights.weights,
                        map_binner.pixel_pointing.detector_pointing.quats,
                    ],
                ).apply(data)

            # Run mapmaking
            self.mapmaker.apply(data)

            # Write
            self.write_splits(data, split_name=split_name)

            # Remove split
            for ob in data.obs:
                spl.remove_split(ob)

        # Restore mapmaker traits
        for k, v in mapmaker_save_traits.items():
            setattr(self.mapmaker, k, v)
        map_binner.pixel_pointing.view = pointing_view_save

    def write_splits(self, data, split_name=None):
        """Write out all split products."""
        if not hasattr(self, "_split_obj"):
            msg = "No splits have been created yet, cannot write"
            raise RuntimeError(msg)

        if split_name is None:
            to_write = dict(self._split_obj)
        else:
            to_write = {split_name: self._split_obj[split_name]}

        if self.mapmaker.write_hdf5:
            fname_suffix = "h5"
        else:
            fname_suffix = "fits"

        for spname, spl in to_write.items():
            mname = f"{self.name}_{split_name}"
            for prod in ["hits", "map", "invcov", "noiseweighted_map"]:
                mkey = f"{mname}_{prod}"
                fname = os.path.join(
                    self.output_dir, f"{self.name}_{spname}_{prod}.{fname_suffix}"
                )
                data[mkey].write(
                    fname,
                    force_serial=self.mapmaker.write_hdf5_serial,
                    single_precision=True,
                )
            # Clean up all products for this split
            for prod in [
                "hits",
                "cov",
                "invcov",
                "rcond",
                "flags",
                "cleaned",
                "binmap",
                "map",
                "noiseweighted_map",
            ]:
                mkey = f"{mname}_{prod}"
                if mkey in data:
                    del data[mkey]

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = {
            "detdata": self.det_data,
        }
        return req

    def _provides(self):
        prov = {"meta": list(), "shared": list(), "detdata": list()}
        return prov

    def _accelerators(self):
        return list()



import numpy as np
from astropy import units as u


"""
A set of functions used in to convert from GPS coordinates to Horizontal coordinates. 
These functions are used in the toast3 drone simulation operator. 

The coordinate systems introduced here are:

- ECEF: Earth Center, Earth Fixed Coordinate System (geocentric coordinates system). 
        It uses a cartesian system centered at the Earth Center-of-Mass to represent
        locations.
- Geodetic: Uses an angular coordinate system (longitude, latitude and geodetic height)
            to describe the Earth. The coordinate system is defined by a datum, which 
            describes the 3D shape of the Earth. The most common datum is the WGS-84, 
            which is also used by GPS and represents the Earth as an Ellipsoid. 
- ENU: Earth-North-Up local coordinate system. It is based on the tangent plane defined 
       by the local vertical direction and the Earth's axis of rotation. The resulting 
       cartesian system has three coordinates: one gives the position along the northern 
       axis (tangent to the meridians), one along the eastern axis (tangent to the paralles)
       and one representing the vertical direction (normal direction to the chosen geodetic
       datum). To maintain the right-hand convention, x is along the East axis, y along 
       the North axis and Z along the Up axis.

Most of the functions are adapted from pymap3d, to include error calulations 
"""

def ellipsoid(model='WGS84'):
    """Return the major and minor semiaxis given an ellipsoid model.
    
    Returns
    -------
    a : astropy.Quantity, semiaxis major in meter
    b : astropy.Quantity, semiaxis minor in meter
    
    """

    if model == 'WGS84':
        a = 6378137.0*u.meter         
        b = 6356752.31424518*u.meter 

    return a, b

def _check_quantity(val, unit):

    if isinstance(val, u.Quantity):
        return val.to(unit)
    else:
        return val * unit

def ecef2lonlat(x, y, z, ell='WGS84'):
    """Convert ECEF to geodetic coordinates

    Based on:  You, Rey-Jer. (2000). 'Transformation of Cartesian to Geodetic
    Coordinates without Iterations'.  Journal of Surveying Engineering. 
    doi: 10.1061/(ASCE)0733-9453

    Parameters
    ----------
    x : float, array (numpy or Quantity)
        target x ECEF coordinate. If numpy float or array, it is considered in meters
    y : float, array (numpy or Quantity)
        target y ECEF coordinate. If numpy float or array, it is considered in meters
    z : float, array (numpy or Quantity)
        target z ECEF coordinate. If numpy float or array, it is considered in meters
    ell : string, optional
        reference ellipsoid
    
    Returns
    -------
    lat : float, array (Quantity)
        target geodetic latitude 
    lon : float, array (Quantity)
        target geodetic longitude
    alt : float, array (Quantity)
        target altitude above geodetic ellipsoid
    
    """

    x = _check_quantity(x, u.meter)
    y = _check_quantity(y, u.meter)
    z = _check_quantity(z, u.meter)

    a, b = ellipsoid(ell)

    r = np.sqrt(x**2 + y**2 + z**2)

    E = np.sqrt(a**2 - b**2)

    # eqn. 4a
    U = np.sqrt(0.5*(r**2 - E**2) + 0.5*np.sqrt((r**2 - E**2)**2 + 4*E**2*z**2))

    Q = np.sqrt(x**2 + y**2)
    huE = np.sqrt(U**2 + E**2)

    Beta = np.arctan2(huE*z, U*Q)

    # eqn. 13
    eps = ((b*U - a*huE + E**2)*np.sin(Beta)) / (
        a*huE*1 / np.cos(Beta) - E**2*np.cos(Beta)
    )
    Beta += eps

    lat = np.arctan2(a*np.tan(Beta), b)

    lon = np.arctan2(y, x)

    # eqn. 7
    alt = np.hypot(z - b * np.sin(Beta), Q - a * np.cos(Beta))

    # inside ellipsoid?
    inside = (x**2/a**2 + y**2/a**2+ z**2/b**2 < 1)

    try:
        if inside.any():  # type: ignore
            # avoid all false assignment bug
            alt[inside] = -alt[inside]
    except (TypeError, AttributeError):
        if inside:
            alt = -alt

    return lon, lat, alt

def hor2enu(az, el, srange, deg=True):
    """Azimuth, Elevation, Slant range to target to East, North, Up
    
    Parameters
    ----------
    azimuth : float, array (numpy or Quantity)
        azimuth, if numpy float or array the unit is determined by the deg parameter
    elevation : float
        elevation, if numpy float or array the unit is determined by the deg parameter
    srange : float, array (numpy or Quantity)
        slant range. If numpy float or array, it is considered in meters
    deg : bool, optional
        if azimuth and elevation are not astropy quantities, if True set them to degrees 
    
    Returns
    --------
    E : Quantity
        East ENU coordinate (meters)
    N : float
        North ENU coordinate (meters)
    U : Quantity
        Up ENU coordinate (meters)
    
    """

    if deg:
        azel_unit = u.deg
    else:
        azel_unit = u.rad

    az = _check_quantity(az, azel_unit)
    el = _check_quantity(el, azel_unit)
    srange = _check_quantity(srange, u.meter)

    r = srange * np.cos(el)

    return r*np.sin(az), r*np.cos(az), srange*np.sin(el)

def enu2ecef(E, N, U, lon, lat, alt, ell='WGS84', deg=True):
    """East, North, Up to ECEF.

    Parameters
    ----------
    E : float, array (numpy or Quantity)
        target east ENU coordinate. If numpy float or array, it is considered in meters
    N : float, array (numpy or Quantity)
        target north ENU coordinate. If numpy float or array, it is considered in meters
    U : float, array (numpy or Quantity)
        target up ENU coordinate. If numpy float or array, it is considered in meters
    lon : float, array (numpy or Quantity)
        geodetic longitude of the observer, if numpy float or array the unit is 
        determined by the deg parameter
    lat : float, array (numpy or Quantity)
        geodetic latitude of the observer, if numpy float or array the unit is 
        determined by the deg parameter 
    alt : float, array (numpy or Quantity)
        altitude above geodetic ellipsoid of the observer, If numpy float or array, 
        it is considered in meters
    ell : string, optional
        reference ellipsoid
    deg : bool, optional
        if lon and lat are not quantities, if True set them to degrees 
    
    Returns
    -------
    x : Quantity
        target x ECEF coordinate
    y : Quantity
        target y ECEF coordinate
    z : Quantity
        target z ECEF coordinate
    
    """

    if deg:
        lonlat_unit = u.deg
    else:
        lonlat_unit = u.rad

    lon = _check_quantity(lon, lonlat_unit)
    lat = _check_quantity(lat, lonlat_unit)
    alt = _check_quantity(alt, u.meter)

    E = _check_quantity(E, u.meter)
    N = _check_quantity(N, u.meter)
    U = _check_quantity(U, u.meter)

    t = np.cos(lat) * U - np.sin(lat) * N
    dz = np.sin(lat) * U + np.cos(lat) * N

    dx = np.cos(lon) * t - np.sin(lon) * E
    dy = np.sin(lon) * t + np.cos(lon) * E

    x0, y0, z0, _, _, _ = lonlat2ecef(lon, lat, alt, ell, deg=deg)

    return x0 + dx, y0 + dy, z0 + dz

def lonlat2ecef(lon, lat, alt, ell='WGS84', deg=True, uncertainty=False, \
                delta_lon=0, delta_lat=0, delta_alt=0):
    """Convert geodetic coordinates to ECEF

    Parameters
    ----------
    lon : float, array (numpy or Quantity)
        geodetic longitude, if numpy float or array the unit is determined by the 
        deg parameter
    lat : float, array (numpy or Quantity)
        geodetic latitude, if numpy float or array the unit is determined by the 
        deg parameter 
    alt : float, array (numpy or Quantity)
        altitude above geodetic ellipsoid, If numpy float or array, it is considered 
        in meters
    ell : string, optional
        reference ellipsoid
    deg : bool, optional
        if azimuth and elevation are not astropy quantities, if True set them to 
        degrees 
    ell : string, optional
        reference ellipsoid
    uncertainty : bool, optional
        if True computes the uncertainties associated in ECEF coordinates
    delta_lon : float, array (numpy or Quantity)
        delta geodetic longitude, if numpy float or array the unit is determined by 
        the deg parameter
    delta_lat : float, array (numpy or Quantity)
        delta geodetic latitude, if numpy float or array the unit is determined by 
        the deg parameter
    delta_alt : float, array (numpy or Quantity)
        delta altitude above geodetic ellipsoid, if numpy float or array the unit is 
        determined by the deg parameter
    
    Returns
    -------
    x : Quantity
        target x ECEF coordinate
    y : Quantity
        target y ECEF coordinate
    z : Quantity
        target z ECEF coordinate
    
    """

    if deg:
        lonlat_unit = u.deg
    else:
        lonlat_unit = u.rad

    lon = _check_quantity(lon, lonlat_unit)
    lat = _check_quantity(lat, lonlat_unit)
    alt = _check_quantity(alt, u.meter)
    
    cos_lat = np.cos(lat)
    sin_lat = np.sin(lat)
    cos_lon = np.cos(lon)
    sin_lon = np.sin(lon)

    a, b = ellipsoid(ell)
        
    N = a**2/np.sqrt(a**2*cos_lat**2 + b**2*sin_lat**2)

    x = (N + alt)*cos_lat*cos_lon
    y = (N + alt)*cos_lat*sin_lon
    z = (N*(b/a)**2 + alt)*sin_lat
    
    if uncertainty:

        delta_lat = _check_quantity(delta_lat, lonlat_unit)
        delta_lon = _check_quantity(delta_lon, lonlat_unit)
        delta_alt = _check_quantity(delta_alt, u.meter)
        
        delta_x, delta_y, delta_z = _lonlat2ecef_error(cos_lon, cos_lat, sin_lon, sin_lat, \
                                                       alt, a, b, delta_lat, delta_lon, delta_alt)
    
    else:
        delta_x, delta_y, delta_z = np.zeros_like(x)*u.meter, \
            np.zeros_like(y)*u.meter, \
            np.zeros_like(z)*u.meter      
    
    return x, y, z, delta_x, delta_y, delta_z

def _lonlat2ecef_error(cos_lon, cos_lat, sin_lon, sin_lat, alt, a, b, \
                       delta_lon, delta_lat, delta_alt):
    """Compute the error in ECEF coordinates.
    
    This uses parameters derived in the lonlat2ecef function. Check that function
    for a deeper explanation.

    """

    delta_x = np.sqrt(delta_alt**2*cos_lon**2*cos_lat**2+\
                      delta_lon**2*(a**2/np.sqrt(a**2*cos_lat**2 + \
                                                 b**2*sin_lat**2) + \
                                    alt)**2*sin_lon**2*cos_lat**2 +\
                      delta_lat**2*(a**2*(a**2*sin_lat*cos_lat - \
                                          b**2*sin_lat*cos_lat)*cos_lon*cos_lat/\
                                    (a**2*cos_lat**2 + b**2*sin_lat**2)**(3/2) - \
                                    (a**2/np.sqrt(a**2*cos_lat**2 + b**2*sin_lat**2) +\
                                    alt)*sin_lat*cos_lon)**2)
        
    delta_y = np.sqrt(delta_alt**2*sin_lon**2*cos_lat**2 + \
                      delta_lon**2*(a**2/np.sqrt(a**2*cos_lat**2 + \
                                                 b**2*sin_lat**2) + \
                                    alt)**2*cos_lon**2*cos_lat**2 + \
                      delta_lat**2*(a**2*(a**2*sin_lat*cos_lat - \
                                          b**2*sin_lat*cos_lat)*sin_lon*cos_lat/\
                                    (a**2*cos_lat**2 + b**2*sin_lat**2)**(3/2) - \
                                    (a**2/np.sqrt(a**2*cos_lat**2 + b**2*sin_lat**2) + \
                                        alt)*sin_lon*sin_lat)**2)
    
    delta_z = np.sqrt(delta_alt**2*sin_lat**2 + \
                      delta_lat**2*(b**2*(a**2*sin_lat*cos_lat - \
                                          b**2*sin_lat*cos_lat)*sin_lat/\
                                    (a**2*cos_lat**2 + b**2*sin_lat**2)**(3/2) +\
                                    (b**2/np.sqrt(a**2*cos_lat**2 + b**2*sin_lat**2) +\
                                        alt)*cos_lat)**2)

    return delta_x, delta_y, delta_z
    
def ecef2enu(ecef_obs_x, ecef_obs_y, ecef_obs_z, \
             ecef_target_x, ecef_target_y, ecef_target_z, \
             delta_obs_x, delta_obs_y, delta_obs_z, \
             delta_target_x, delta_target_y, delta_target_z, \
             lon, lat, deg=True):
    """Return the position relative to a refence point in a ENU system. 
    
    If one of delta_obs or delta_target are different from zeros then uncertainties is 
    calculated automatically. The array returns as EAST, NORTH, UP
    
    Parameters
    ----------
    ecef_obs_x : float or array, numpy or Quantity
        x ECEF coordinates array of the observer, if numpy float or array the unit
        is meter
    ecef_obs_y : float or array, numpy or Quantity
        y ECEF coordinates array of the observer, if numpy float or array the unit 
        is meter
    ecef_obs_z : float or array, numpy or Quantity
        z ECEF coordinates array of the observer, if numpy float or array the unit 
        is meter
    ecef_target_x : float or array, numpy or Quantity
        x ECEF coordinates array of the target, if numpy float or array the unit 
        is meter
    ecef_target_y : float or array, numpy or Quantity
        y ECEF coordinates array of the target, if numpy float or array the unit 
        is meter
    ecef_target_z : float or array, numpy or Quantity
        z ECEF coordinates array of the target, if numpy float or array the unit 
        is meter
    delta_obs_x : numpy or Quantity, numpy or Quantity
        x ECEF Coordinates error of the observer, if numpy array the unit is meter
    delta_obs_y : numpy or Quantity, numpy or Quantity
        x ECEF Coordinates error of the observer, if numpy array the unit is meter
    delta_obs_z : numpy or Quantity, numpy or Quantity
        x ECEF Coordinates error of the observer, if numpy array the unit is meter
    delta_target_x : numpy or Quantity, numpy or Quantity
        x ECEF Coordinates error of the target, if numpy array the unit is meter
    delta_target_y : numpy or Quantity, numpy or Quantity
        y ECEF Coordinates error of the target, if numpy array the unit is meter
    delta_target_z : numpy or Quantity, numpy or Quantity
        z ECEF Coordinates error of the target, if numpy array the unit is meter
    lon : float, array (numpy or Quantity)
        geodetic longitude, if numpy float or array the unit is determined by the deg parameter
    lat : float, array (numpy or Quantity)
        geodetic latitude, if numpy float or array the unit is determined by the deg parameter 
    deg : bool, optional
        if azimuth and elevation are not astropy quantities, if True set them to degrees

    Returns
    --------
    E : Quantity
        East ENU coordinate error
    N : float
        North ENU coordinate error
    U : Quantity
        Up ENU coordinate error
    delta_E : Quantity
        East ENU coordinate error
    delta_N : float
        North ENU coordinate error
    delta_U : Quantity
        Up ENU coordinate error
    
    """
    
    if deg:
        lonlat_unit = u.deg
    else:
        lonlat_unit = u.rad

    lon = _check_quantity(lon, lonlat_unit)
    lat = _check_quantity(lat, lonlat_unit)

    ecef_obs_x = _check_quantity(ecef_obs_x, u.meter)
    ecef_obs_y = _check_quantity(ecef_obs_y, u.meter)
    ecef_obs_z = _check_quantity(ecef_obs_z, u.meter)

    ecef_target_x = _check_quantity(ecef_target_x, u.meter)
    ecef_target_y = _check_quantity(ecef_target_y, u.meter)
    ecef_target_z = _check_quantity(ecef_target_z, u.meter)

    delta_obs_x = _check_quantity(delta_obs_x, u.meter)
    delta_obs_y = _check_quantity(delta_obs_y, u.meter)
    delta_obs_z = _check_quantity(delta_obs_z, u.meter)

    delta_target_x = _check_quantity(delta_target_x, u.meter)
    delta_target_y = _check_quantity(delta_target_y, u.meter)
    delta_target_z = _check_quantity(delta_target_z, u.meter)
    
    cos_lat = np.cos(lat)
    sin_lat = np.sin(lat)
    cos_lon = np.cos(lon)
    sin_lon = np.sin(lon)

    mat = np.array([[-sin_lon, cos_lon, 0],
                    [-sin_lat*cos_lon, -sin_lat*sin_lon, cos_lat],
                    [cos_lat*cos_lon, cos_lat*sin_lon, sin_lat]])

    ecef_target = np.vstack((ecef_target_x, ecef_target_y, ecef_target_z))
    ecef_obs = np.vstack((ecef_obs_x, ecef_obs_y, ecef_obs_z))
 
    diff = ecef_target-ecef_obs
    enu = np.matmul(mat, diff)

    delta_obs = np.vstack((delta_obs_x, delta_obs_y, delta_obs_z))
    delta_target = np.vstack((delta_target_x, delta_target_y, delta_target_z))

    if np.any(delta_obs != 0) or np.any(delta_target != 0):
        delta_diff = np.sqrt(delta_obs**2+delta_target**2)
        delta_enu = np.matmul(mat, delta_diff)
    else:
        delta_enu = np.zeros_like(enu)
        delta_enu = _check_quantity(delta_enu, u.meter)
    
    return enu[0], enu[1], enu[2], delta_enu[0], delta_enu[1], delta_enu[2]

def enu2hor(E, N, U, delta_E, delta_N, delta_U, degrees=True):
    """Compute Horizontal coordinates given the coordinates in the East-North-Up system.

    Parameters
    ----------
    E : float, array (numpy or Quantity)
        target east ENU coordinate. If numpy float or array, it is considered in meters
    N : float, array (numpy or Quantity)
        target north ENU coordinate. If numpy float or array, it is considered in meters
    U : float, array (numpy or Quantity)
        target up ENU coordinate. If numpy float or array, it is considered in meters

    """

    E = _check_quantity(E, u.meter)
    N = _check_quantity(N, u.meter)
    U = _check_quantity(U, u.meter)

    delta_E = _check_quantity(delta_E, u.meter)
    delta_N = _check_quantity(delta_N, u.meter)
    delta_U = _check_quantity(delta_U, u.meter)

    s = np.sqrt(E**2+N**2+U**2)
    el = np.arctan2(U, np.sqrt(E**2+N**2))
    az = np.arctan2(E, N) % (2*np.pi*u.rad)

    if np.any(delta_E != 0) or np.any(delta_N != 0) or np.any(delta_U != 0):
    
        delta_s = np.sqrt((E**2*delta_E**2+N**2*delta_N**2+U**2*delta_U**2)/s**2)

        delta_el = np.sqrt((delta_U**2*(E**2 + N**2)**2 + \
                            U**2*(delta_E**2*E**2 + delta_N**2*N**2))/\
                           ((E**2 + N**2)*s**4))*el.unit
        
        delta_az = np.sqrt((delta_E**2*N**2 + delta_N**2*E**2)/(E**2 + N**2)**2)*az.unit
    
    else:
        delta_s = np.zeros_like(s)*u.meter
        delta_az = np.zeros_like(az)*az.unit
        delta_el = np.zeros_like(el)*el.unit

    return az, el, s, delta_az, delta_el, delta_s



# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Check that pointing expanded with TOAST is compatible with the MLMapmaker

"""

import os
import tempfile
import unittest

import astropy.units as u
import numpy as np
from pixell import utils, enmap, curvedsky

try:
    # Import sotodlib toast module first, which sets global toast defaults
    import sotodlib.toast as sotoast
    import sotodlib.toast.ops as so_ops
    import toast
    from toast.observation import default_values as defaults
    toast_available = True
except ImportError as e:
    toast_available = False

if toast_available:
    import healpy as hp

from . import _helpers as helpers


class MapmakerPointingTest(unittest.TestCase):
    def test_mapmaker_pointing(self):
        if not toast_available:
            print("toast cannot be imported- skipping unit tests", flush=True)
            return

        log = toast.utils.Logger.get()

        nnz = 3
        nside = 1024
        wpix = hp.nside2pixarea(nside, degrees=True)**.5 * 60
        npix = 12 * nside**2
        testdir = tempfile.TemporaryDirectory()

        comm, procs, rank = toast.get_world()
        data = helpers.simulation_test_data(
            comm,
            telescope_name=None,
            wafer_slot="w00",
            bands="LAT_f230",
            sample_rate=10.0 * u.Hz,
            thin_fp=16,
            cal_schedule=False,
        )

        pointing = toast.ops.PointingDetectorSimple(
            name="det_pointing_radec",
            quats="quats_radec",
            boresight=defaults.boresight_radec,
            shared_flag_mask=0,
        )
        pixels = toast.ops.PixelsHealpix(
            nside=nside,
            create_dist="pixel_dist",
            detector_pointing=pointing,
        )
        weights = toast.ops.StokesWeights(
            name="weights_radec",
            weights="weights_radec",
            mode="IQU",
            detector_pointing=pointing,
        )

        input_map_file = os.path.join(
            testdir.name,
            "pointing_test_map.fits",
        )
        if rank == 0:
            input_map = np.zeros([nnz, npix], dtype=np.float32)
            input_map[1] = 1
            hp.write_map(
                input_map_file, input_map, nest=pixels.nest, column_units="K"
            )
        log.info_rank(f"Wrote test map to {input_map_file}", comm=comm)

        if data.comm.comm_world is not None:
            data.comm.comm_world.Barrier()

        # Scan map into timestreams
        scan_hpix = toast.ops.ScanHealpixMap(
            file=input_map_file,
            det_data=defaults.det_data,
            pixel_pointing=pixels,
            stokes_weights=weights,
        )
        scan_hpix.apply(data)

        # Bin the map using TOAST
        toast.ops.DefaultNoiseModel(noise_model="noise_model").apply(data)
        binner = toast.ops.BinMap(
            pixel_dist="pixel_dist",
            pixel_pointing=pixels,
            stokes_weights=weights,
            noise_model="noise_model",
            shared_flag_mask=0,
        )
        toast.ops.MapMaker(
            name="mapmaker",
            det_data=defaults.det_data,
            binning=binner,
            write_hits=True,
            write_binmap=True,
            write_rcond=False,
            write_invcov=False,
            write_cov=False,
            write_map=False,
            output_dir=testdir.name,
        ).apply(data)
        file_hits_toast = os.path.join(testdir.name, "mapmaker_hits.fits")
        file_map_toast = os.path.join(testdir.name, "mapmaker_binmap.fits")

        # Create an area file
        areafile = os.path.join(testdir.name, "area_file.fits")
        box = utils.parse_box("-51:-35,35:49") * utils.degree
        res = wpix * utils.arcmin
        shape, wcs = enmap.geometry(box, res=res, proj="car")
        wcs.wcs.crpix[1] -= 0.5
        enmap.write_map_geometry(areafile, shape, wcs)

        mapmaker = so_ops.MLMapmaker(
            area=areafile,
            name="mlmapmaker",
            out_dir=testdir.name,
            comps="TQU",
            nmat_type="Nmat",
            maxiter=[3],
            weather="typical",
            truncate_tod=False,
            write_hits=True,
            write_rhs=False,
            write_div="all",
            write_bin=True,
            deslope=False,
        )
        mapmaker.apply(data)

        if rank == 0:
            # Direct comparison of pointing
            obs = data.obs[0]
            pmap = mapmaker.signal_map.data[obs.name].pmap
            fplane = pmap._get_asm().fplane
            coords = np.array(pmap.sight.coords(fplane))
            dets = mapmaker.mapmaker.data[0].dets
            ndet = len(dets)

            pointing.apply(data)
            dlats_mean = np.zeros(ndet)
            dlats_rms = np.zeros(ndet)
            dlons_mean = np.zeros(ndet)
            dlons_rms = np.zeros(ndet)
            dgammas_mean = np.zeros(ndet)
            dgammas_rms = np.zeros(ndet)
            for idet in range(ndet):
                lon, lat, cosgamma, singamma = coords[idet].T
                gamma = np.arctan2(singamma, cosgamma)
                theta, toast_lon, toast_gamma = toast.qarray.to_iso_angles(
                    obs.detdata["quats_radec"][idet]
                )
                toast_lat = np.pi / 2 - theta
                toast_dets = obs.local_detectors
                dlon = np.degrees(lon - toast_lon) * 3600
                dlat = np.degrees(lat - toast_lat) * 3600
                dgamma = np.degrees(np.unwrap(gamma - toast_gamma)) * 3600
                mean_dlon = np.mean(dlon)
                rms_dlon = np.std(dlon)
                mean_dlat = np.mean(dlat)
                rms_dlat = np.std(dlat)
                mean_dgamma = np.mean(dgamma)
                rms_dgamma = np.std(dgamma)
                #log.info(f"dlat = {mean_dlat:.3f} +- {rms_dlat:.3f} arcsec")
                #log.info(f"dlon = {mean_dlon:.3f} +- {rms_dlon:.3f} arcsec")
                #log.info(f"dgamma = {mean_dgamma:.3f} +- {rms_dgamma:.3f} arcsec")
                dlats_mean[idet] = mean_dlat
                dlats_rms[idet] = rms_dlat
                dlons_mean[idet] = mean_dlon
                dlons_rms[idet] = rms_dlon
                dgammas_mean[idet] = mean_dgamma
                dgammas_rms[idet] = rms_dgamma
            log.info(f"dlat = {np.mean(dlats_mean):.3f} +- {np.std(dlats_mean):.3f} arcsec")
            log.info(f"dlon = {np.mean(dlons_mean):.3f} +- {np.std(dlons_mean):.3f} arcsec")
            log.info(f"dgamma = {np.mean(dgammas_mean):.3f} +- {np.std(dgammas_mean):.3f} arcsec")
            tol = 1.0
            if rms_dlon > tol:
                raise RuntimeError(f"RA differs systematically over {tol} arc seconds")
            if rms_dlat > tol:
                raise RuntimeError(f"Dec differs systematically over {tol} arc seconds")
            if rms_dgamma > tol:
                raise RuntimeError(f"PA differs systematically over {tol} arc seconds")

            # Compare binned maps

            file_hits = os.path.join(testdir.name, "mlmapmaker_sky_hits.fits")
            file_map = os.path.join(testdir.name, "mlmapmaker_sky_map.fits")
            hits = enmap.read_map(file_hits).astype(int)
            sky = enmap.read_map(file_map)
            good = hits != 0
            ind = sky[1] != 0

            toast_hits = hp.read_map(file_hits_toast)
            toast_sky = hp.read_map(file_map_toast, None)
            toast_good = toast_hits != 0
            toast_ind = toast_sky[1] != 0

            rms = []
            means = []
            toast_rms = []
            toast_means = []
            log.info(f"            {'MLMapmaker':16}            {'TOAST':16}")
            log.info(f"Hits        {np.sum(hits):8}       {np.sum(toast_hits):16}")
            log.info(f"Hit pixels  {np.sum(good):8}       {np.sum(toast_good):16}")
            log.info(f"After cut   {np.sum(ind):8}       {np.sum(toast_ind):16}")
            for i in range(3):
                rms.append(np.std(sky[i][ind].ravel()))
                means.append(np.mean(sky[i][ind].ravel()))
                toast_rms.append(np.std(toast_sky[i][toast_ind].ravel()))
                toast_means.append(np.mean(toast_sky[i][toast_ind].ravel()))
                stokes = "IQU"[i]
                log.info(
                    f"{stokes:3} {means[i]:12.6f} +- {rms[i]:12.6f}, "
                    f"{toast_means[i]:12.6f} +- {toast_rms[i]:12.6f}"
                )

            tol = 1e-3
            if np.abs(means[0]) > tol:
                raise RuntimeError("Found non-zero I")

            if np.abs(means[1] - 1) > tol:
                raise RuntimeError("Found non-unit Q")

            if np.abs(means[2]) > tol:
                raise RuntimeError("Found non-zero U")

        helpers.close_data_and_comm(data)

if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import numpy as np
from .. import core
from .helpers import _valid_arg
from so3g.proj import quat

import logging
logger = logging.getLogger(__name__)

DEG = np.pi / 180


def apply_basic_pointing_model(tod):
    """
    Updates and returns the tod.boresight AxisManager after applying pointing model parameters.

    Args:
        tod: AxisManager (must contain 'ancil' field)

    Returns:
        _boresight : AxisManager containing new boresight data.
    """

    _ancil = _valid_arg("ancil", src=tod)
    telescope = _valid_arg("obs_info", src=tod)["telescope"]
    # Start with a fresh axismanager
    _boresight = core.AxisManager(tod.samps)

    if "lat" in telescope:
        # Returns in Radians
        _boresight.wrap("az", _ancil.az_enc * DEG, [(0, "samps")])
        _boresight.wrap("el", _ancil.el_enc * DEG, [(0, "samps")])
        _boresight.wrap(
            "roll", (_ancil.el_enc - 60 - _ancil.corotator_enc) * DEG, [(0, "samps")]
        )

    elif "sat" in telescope:
        # Returns in radians
        _boresight.wrap("az", _ancil.az_enc * DEG, [(0, "samps")])
        _boresight.wrap("el", _ancil.el_enc * DEG, [(0, "samps")])
        _boresight.wrap("roll", -1 * _ancil.boresight_enc * DEG, [(0, "samps")])

    if 'boresight' in tod:
        del tod['boresight']

    tod.wrap('boresight', _boresight)

    return _boresight


def apply_pointing_model(tod, pointing_model=None, ancil=None,
                         wrap=None):
    """Applies a static pointing model to compute corrected boresight
    position and orientation in horizon coordinates.  The encoder
    values in tod.ancil are consumed as raw data, and the computed
    values are stored in tod.boresight.

    Args:
      tod (AxisManager): the observation data.
      pointing_model (AxisManager): if None, the pointing_model
        parameters are read from tod.pointing_model.
      ancil (AxisManager): if None, the encoders are read from
        tod.ancil.
      wrap (str): If specified, the name in tod where corrected
        boresight should be stored.  If None, the default of
        'boresight' is used.  Pass wrap=False to not store the
        result in tod.

    Returns:
      AxisManager: the corrected boresight.

    """
    if pointing_model is None and "pointing_model" not in tod:
        logger.warning("No pointing_model found -- applying basic model.")
        assert wrap in (
            None,
            "boresight",
        ), "When using naive pointing model, wrap=... not supported"
        return apply_basic_pointing_model(tod)

    pointing_model = _valid_arg(pointing_model, "pointing_model", src=tod)
    ancil = _valid_arg(ancil, "ancil", src=tod)

    # Encoder values, to radians.
    if pointing_model is None:
        raise ValueError("No pointing_model specified")
    if "version" not in pointing_model:
        raise ValueError("Pointing model version not specified")
    vers = pointing_model["version"]
    tel_type = vers.split("_")[0]
    if tel_type == "sat":
        boresight = apply_pointing_model_sat(vers, pointing_model, tod, ancil)
    elif tel_type == "lat":
        boresight = apply_pointing_model_lat(vers, pointing_model, ancil)
    else:
        raise ValueError(f'Unimplemented pointing model "{vers}"')

    if wrap is None:
        wrap = "boresight"
    if wrap is not False:
        if wrap in tod._fields:
            del tod[wrap]
        tod.wrap(wrap, boresight)
    return boresight


def apply_pointing_model_sat(vers, params, tod, ancil):
    az, el, roll = _get_sat_enc_radians(ancil)

    if vers == 'sat_naive':
        return _new_boresight(ancil.samps, az=az, el=el, roll=roll)

    elif vers == 'sat_v1':
        az1, el1, roll1 = model_sat_v1(params, az, el, roll)
        return _new_boresight(ancil.samps, az=az1, el=el1, roll=roll1)

    else:
        raise ValueError(f'Unimplemented pointing model "{vers}"')


def apply_pointing_model_lat(vers, params, ancil):
    az, el, roll = _get_lat_enc_radians(ancil)
    
    if vers == 'lat_naive':
        return _new_boresight(ancil.samps, az=az, el=el, roll=roll)
        
    if vers == "lat_v1":
        az1, el1, roll1 = model_lat_v1(params, az, el, roll)
        return _new_boresight(ancil.samps, az=az1, el=el1, roll=roll1)

    else:
        raise ValueError(f'Unimplemented pointing model "{vers}"')


#
# LAT model(s)
#
def model_lat_v1(params, az, el, roll):
    """Applies pointing model to (az, el, roll).

    Args:
      params: AxisManager (or dict) of pointing parameters.
      az, el, roll: naive horizon coordinates, in radians, of the
        boresight.

    The implemented model parameters are all in radians:
    - enc_offset_{az, el, cr}: Encoder offsets in radians.
      Sign convention: True = Encoder + Offset
    - cr_center_{xi,eta}0: The (xi,eta) coordinate in the LATR-centered 
      focal plane that remains fixed under corotation. 
    - el_axis_center_{xi,eta}0: The (xi,eta) coordinate in the CR-centered
      focal plane that appears fixed when the elevation structure is rotated 
      about its axis.
    - mir_center_{xi,eta}0: The (xi,eta) coordinate in the El-structure-centered
      focal plane that appears fixed when the mirrors are rotated about the ray from
      sky that hits the center of both mirrors.

    """
    _p = dict(param_defaults['lat_v1'])
    if isinstance(params, dict):
        _p.update(params)
    else:
        _p.update({k: params[k] for k in params._fields.keys()})
    params, _p = _p, None   
         
    for k, v in params.items():
        if k == 'version':
            continue
        if k not in param_defaults['lat_v1'] and v != 0.:
            raise ValueError(f'Handling of model param "{k}" is not implemented.')
    
    # Here we reconstruct the naive corotator value before applying 
    # the elevation encoder offset.
    cr = el - roll - np.deg2rad(60)    
    
    az_orig = az.copy()
    az = az + params['enc_offset_az']
    el = el + params['enc_offset_el'] 
    cr = cr + params['enc_offset_cr']
    
    # Lonlat rotation with az and el encoder offsets included.
    q_lonlat = quat.rotation_lonlat(-1 * az, el)

    # Rotation that takes a vector in elevation-hub centered 
    # coordinates to final boresight centered coordinates. 
    # Accounts for any mirror-related offsets.
    q_mir_center = ~quat.rotation_xieta(
        params['mir_center_xi0'],
        params['mir_center_eta0']
    )

    # Elevation component of roll motion
    q_el_roll = quat.euler(2, el - np.deg2rad(60))
    
    # Rotation that takes a vector in telescope's corotator-centered
    # coordinates to el-hub-centered coordinates
    q_el_axis_center = ~quat.rotation_xieta(
        params['el_axis_center_xi0'],
        params['el_axis_center_eta0']
    )

    # Corotator component of roll motion
    q_cr_roll = quat.euler(2, -1 * cr)
    
    # Rotation that takes a vector in LATR/focal plane-centered coordinates to  
    # corotator-centered coordinates.
    q_cr_center = ~quat.rotation_xieta(
        params['cr_center_xi0'],
        params['cr_center_eta0']
    )

    # Horizon Coordinates
    q_hs = (
        q_lonlat * q_mir_center
        * q_el_roll * q_el_axis_center
        * q_cr_roll * q_cr_center
    )
    new_az, el, roll = quat.decompose_lonlat(q_hs)* np.array([-1, 1, 1])[..., None]
    
    # Make corrected az as close as possible to the input az.    
    change = ((new_az - az_orig) + np.pi) % (2 * np.pi) - np.pi
    az = az_orig + change

    return az, el, roll


#
# SAT model(s)
#

# sat_v1: you can expand v1, as long as new params don't do anything
# if their value is zero (and that should be the registered default).

def model_sat_v1(params, az, el, roll):
    """Applies pointing model to (az, el, roll).

    Args:
      params: AxisManager (or dict) of pointing parameters.
      az, el, roll: naive horizon coordinates, in radians, of the
        boresight.

    The implemented model parameters are:

      - fp_rot_{xi,eta}0: within the focal plane (i.e. relative to the
        corrected boresight), the center of rotation of the boresight.
        Radians.
      - fp_offset_{xi,eta}0: Offset of focal plane's rotational center 
        relative to position in focal plane that lies along the optical axis.
        Corrects for "collimation error".
      - enc_offset_{az,el,boresight}: Encoder offsets, in Radians. 
        Sign convention: True = Encoder + Offset
      - base_tilt_{cos,sin}: Base tilt coefficients, in radians. 
      - az_rot: Dimensionless parameter describing a linear dependence of Az on El.

    """
    _p = dict(param_defaults['sat_v1'])
    if isinstance(params, dict):
        _p.update(params)
    else:
        _p.update({k: params[k] for k in params._fields.keys()})
    params, _p = _p, None

    for k, v in params.items():
        if k == 'version':
            continue
        if k not in param_defaults['sat_v1'] and v != 0.:
            raise ValueError(f'Handling of model param "{k}" is not implemented.')

    # Construct offsetted encoders.
    az_orig = az.copy()
    az_twist = params['az_rot'] * (el + params['enc_offset_el'])
    az = az + params['enc_offset_az'] + az_twist
    el = el + params['enc_offset_el'] 
    roll = roll - params['enc_offset_boresight'] 

    # Rotation that tilts the base (referred to vals after enc correction).
    base_tilt = get_base_tilt_q(params['base_tilt_cos'], params['base_tilt_sin'])

    # Rotation that takes a vector in array-centered focal plane coords
    # to a vector in boresight-rotation-centered focal plane coords.
    q_fp_rot = ~quat.rotation_xieta(params['fp_rot_xi0'], params['fp_rot_eta0'])
    
    # Rotation that moves the center of the focal plane to fp_offset_(xi, eta)0.  
    q_fp_offset = quat.rotation_xieta(params['fp_offset_xi0'], params['fp_offset_eta0'])

    # Horizon coordinates.
    q_hs = (base_tilt * quat.rotation_lonlat(-az, el)
            * q_fp_offset * ~q_fp_rot * quat.euler(2, roll) * q_fp_rot)

    neg_az, el, roll = quat.decompose_lonlat(q_hs)

    # Make corrected az as close as possible to the input az.
    change = ((-neg_az - az_orig) + np.pi) % (2 * np.pi) - np.pi
    new_az = az_orig + change

    return new_az, el, roll


# Support functions
param_defaults={
    'lat_v1' : {
        'enc_offset_az': 0,
        'enc_offset_el': 0,
        'enc_offset_cr': 0,
        'el_axis_center_xi0': 0,
        'el_axis_center_eta0': 0,
        'cr_center_xi0': 0,
        'cr_center_eta0': 0,
        'mir_center_xi0': 0,
        'mir_center_eta0': 0,
    },
    'sat_v1' : {
        'enc_offset_az': 0.,
        'enc_offset_el': 0.,
        'enc_offset_boresight': 0.,
        'fp_offset_xi0': 0.,
        'fp_offset_eta0': 0.,
        'fp_rot_xi0': 0.,
        'fp_rot_eta0': 0.,
        'az_rot': 0.,
        'base_tilt_cos': 0.,
        'base_tilt_sin': 0.
    }
}

def _new_boresight(samps, az=None, el=None, roll=None):
    boresight = core.AxisManager(samps)
    for k, v in zip(['az', 'el', 'roll'], [az, el, roll]):
        boresight.wrap_new(k, shape=('samps',), dtype='float64')
        if v is not None:
            boresight[k][:] = v
    return boresight

def _get_sat_enc_radians(ancil):
    return (ancil.az_enc * DEG,
            ancil.el_enc * DEG,
            -ancil.boresight_enc * DEG)

def _get_lat_enc_radians(ancil):
    az = ancil.az_enc * DEG
    el = ancil.el_enc * DEG
    if "roll_enc" in ancil:
        roll = ancil.roll_enc * DEG
    else:
        roll = (ancil.el_enc - ancil.corotator_enc - 60) * DEG
    return (az, el, roll)

def get_base_tilt_q(c, s):
    """Returns the quaternion rotation that applies base tilt, taking
    vectors in the platforms horizon coordinates to vectors in the
    site's local horizon coordinates.  The c and s parameters together
    define a direction and amplitude of the base tilt.

    In this implementation, c and s have the same meaning and sign
    convention as TPOINT parameters AN and AW, respectively.

    """
    # Imagine az=-phi
    phi = np.arctan2(s, c)
    # And that base tilt causes the true el to lie below the expected
    # (encoder) el, at that position.
    amp = (c**2 + s**2)**.5
    return quat.euler(2, phi) * quat.euler(1, amp) * quat.euler(2, -phi)




import numpy as np
from scipy.optimize import curve_fit
from lmfit import Model as LmfitModel
from sotodlib import core, tod_ops
from sotodlib.tod_ops import filters, apodize, fft_ops
import logging

logger = logging.getLogger(__name__)


def get_hwpss(aman, signal=None, hwp_angle=None, bin_signal=True, bins=360,
              lin_reg=True, modes=[1, 2, 3, 4, 5, 6, 7, 8], apply_prefilt=True,
              prefilt_cfg=None, prefilt_detrend='linear', flags=None,
              apodize_edges=True, apodize_edges_samps=1600, 
              apodize_flags=True, apodize_flags_samps=200,
              merge_stats=True, hwpss_stats_name='hwpss_stats',
              merge_model=True, hwpss_model_name='hwpss_model'):
    """
    Extracts HWP synchronous signal (HWPSS) from a time-ordered data (TOD)
    using linear regression or curve-fitting. The curve-fitting or linear
    regression are either run on the full time ordered data vs hwp angle
    or the time ordered data binned in hwp_angle. If the curve-fitting
    option is used it must be performed on the binned data.

    Parameters
    ----------
    aman : AxisManager object
        The TOD to extract HWPSS from.
    signal : str or None
        The field name in the axis manager to use for the TOD signal.
        If not provided, ``signal`` will be used.
    hwp_angle : array-like, optional
        The HWP angle for each sample in `aman`. If not provided, `aman.hwp_angle` will be used.
    bin_signal : bool, optional
        Whether to bin the TOD signal into HWP angle bins before extracting HWPSS. Default is `True`.
    bins : int, optional
        The number of HWP angle bins to use if `bin_signal` is `True`. Default is 360.
    lin_reg : bool, optional
        Whether to use linear regression to extract HWPSS from the binned signal. If `False`, curve-fitting will be used instead.
        Default is `True`.
    modes : list of int, optional
        The HWPSS harmonic modes to extract. Default is [1, 2, 3, 4, 5, 6, 7, 8].
    apply_prefilt : bool, optional
        Whether to apply a high-pass filter to signal before extracting HWPSS. Default is `True`.
        If run through preprocess and `signal` is not `aman.signal` then default to `False`.
    prefilt_cfg : dict, optional
        The configuration of the high-pass filter, in Hz. Only used if `apply_prefilt` is `True`.
        Default is sine2 filter of with cutoff frequency of 1.0 Hz and trans_width of 1.0 Hz.
    prefilt_detrend: str or None
        Method of detrending when you apply prefilter. Default is `linear`. If data is already detrended or you do not want to detrend,
        set it to `None`.
    flags : str or RangesMatrix or Ranges, optional
        Flags to be masked out before extracting HWPSS. If Default is None, and no mask will be applied.
        If provided by a string, `aman.flags.get(flags)` is used for the flags.
    merge_stats : bool, optional
        Whether to add the extracted HWPSS statistics to `aman` as new axes. Default is `True`.
    hwpss_stats_name : str, optional
        The name to use for the new field containing the HWPSS statistics if `merge_stats` is `True`. Default is 'hwpss_stats'.
    merge_extract : bool, optional
        Whether to add the extracted HWPSS to `aman` as a new signal field. Default is `True`.
    hwpss_extract_name : str, optional
        The name to use for the new signal field containing the extracted HWPSS if `merge_extract` is `True`. Default is 'hwpss_extract'.

    Returns
    -------
    hwpss_stats : AxisManager object
        The extracted HWPSS and its statistics. The statistics include:

            - **coeffs** (n_dets x n_modes) : coefficients of the model

            .. math::
                \sum_n \mathrm{coeffs}[2n]\sin{(\mathrm{modes}[n] \chi_{\mathrm{hwp}})} + \mathrm{coeffs}[2n+1]\cos{(\mathrm{modes}[n] \chi_{\mathrm{hwp}})}

            where the sum on n range(len(modes)). **Note**: n_modes is 2*len(modes)

            - **covars** (n_dets x n_modes x n_modes) : variance covariance matrix of the fitted coefficients for each detector.
            - **redchi2** (n_dets) : reduced chi^2 of the fit for each detector.
        
            **In the binned case the following are returned:**
            
            - **binned_angle** (n_bins) : binned version of hwp_angle in range (0, 2pi] with number of bins set by bins argument.
            - **bin_counts** (n_dets x n_bins): sample counts of each bin for each detector.
            - **binned_signal** (n_dets x n_bins) : binned signal for each detector.
            - **sigma_bin** (n_dets) : average over all bins of the standard deviation of the signal within each bin.
        
            **In the non-binned case the following are returned:**

            - **sigma_tod** (n_dets) : estimate of the standard deviation of the signal using function ``estimate_sigma_tod``
    """
    if prefilt_cfg is None:
        prefilt_cfg = {'type': 'sine2', 'cutoff': 1.0, 'trans_width': 1.0}

    prefilt = filters.get_hpf(prefilt_cfg)
    if signal is None:
        #signal_name variable to be deleted when tod_ops.fourier_filter is updated
        signal_name = 'signal'
        signal = aman[signal_name]
    elif isinstance(signal, str):
        signal_name = signal
        signal = aman[signal_name]
    elif isinstance(signal, np.ndarray):
        raise TypeError("Currently ndarray not supported, need update to tod_ops.fourier_filter module to remove signal_name argument.")
    else:
        raise TypeError("Signal must be None, str, or ndarray")

    if apply_prefilt:
        # This requires signal to be a string.
        signal = np.array(tod_ops.fourier_filter(
                aman, prefilt, detrend=prefilt_detrend, signal_name=signal_name)
                )

    if hwp_angle is None:
        hwp_angle = aman.hwp_angle

    # define hwpss_stats
    mode_names = []
    for mode in modes:
        mode_names.append(f'S{mode}')
        mode_names.append(f'C{mode}')

    hwpss_stats = core.AxisManager(aman.dets, core.LabelAxis(
        name='modes', vals=np.array(mode_names, dtype='<U3')))
    if bin_signal:
        hwp_angle_bin_centers, bin_counts, binned_hwpss, hwpss_sigma_bin = get_binned_hwpss(
            aman, signal, hwp_angle=None, bins=bins, flags=flags, 
            apodize_edges=apodize_edges, apodize_edges_samps=apodize_edges_samps, 
            apodize_flags=apodize_flags, apodize_flags_samps=apodize_flags_samps,)
        
        # check bin count
        num_invalid_bins = np.count_nonzero(np.isnan(binned_hwpss[0][:]))
        if num_invalid_bins > 0:
            logger.warning(f'There are {num_invalid_bins} bins with zero samples. ' + 
                             'You maybe using simulation data whose hwp speed is perfectly constant, ' + 
                             'or your specification of number of bins is too large.')
        
        # wrap
        hwpss_stats.wrap('binned_angle', hwp_angle_bin_centers, [
                       (0, core.IndexAxis('bin_hwp_samps', count=bins))])
        hwpss_stats.wrap('bin_counts', bin_counts, [
                       (0, 'dets'), (1, 'bin_hwp_samps')])
        hwpss_stats.wrap('binned_signal', binned_hwpss, [
                       (0, 'dets'), (1, 'bin_hwp_samps')])
        hwpss_stats.wrap('sigma_bin', hwpss_sigma_bin, [(0, 'dets')])

        if lin_reg:
            fitsig_binned, coeffs, covars, redchi2s = hwpss_linreg(
                x=hwp_angle_bin_centers, ys=binned_hwpss, yerrs=hwpss_sigma_bin, modes=modes)
        else:
            params_init = guess_hwpss_params(
                x=hwp_angle_bin_centers, ys=binned_hwpss, modes=modes)
            fitsig_binned, coeffs, covars, redchi2s = hwpss_curvefit(x=hwp_angle_bin_centers, ys=binned_hwpss, yerrs=hwpss_sigma_bin,
                                                                     modes=modes, params_init=params_init)
        # tod template
        fitsig_tod = harms_func(hwp_angle, modes, coeffs)

        # wrap the optimal values and stats
        hwpss_stats.wrap('binned_model', fitsig_binned,
                       [(0, 'dets'), (1, 'bin_hwp_samps')])
        hwpss_stats.wrap('coeffs', coeffs, [(0, 'dets'), (1, 'modes')])
        hwpss_stats.wrap('covars', covars, [
                       (0, 'dets'), (1, 'modes'), (2, 'modes')])
        hwpss_stats.wrap('redchi2s', redchi2s, [(0, 'dets')])

    else:
        if isinstance(flags, str):
            flags = aman.flags.get(flags)
        if flags is None:
            m = np.ones([aman.dets.count, aman.samps.count], dtype=bool)
        else:
            m = ~flags.mask()

        hwpss_sigma_tod = estimate_sigma_tod(signal, hwp_angle)
        hwpss_stats.wrap('sigma_tod', hwpss_sigma_tod, [(0, 'dets')])

        if lin_reg:
            fitsig_tod, coeffs, covars, redchi2s = hwpss_linreg(
                x=hwp_angle, ys=signal, yerrs=hwpss_sigma_tod, modes=modes)

        else:
            raise ValueError('Curve-fitting for TOD are specified.' +
                             'It will take too long time and return meaningless result.' +
                             'Specify (bin_signal, lin_reg) = (True, True) or (True, False) or (False, True)')

        hwpss_stats.wrap('coeffs', coeffs, [(0, 'dets'), (1, 'modes')])
        hwpss_stats.wrap('covars', covars, [
                       (0, 'dets'), (1, 'modes'), (2, 'modes')])
        hwpss_stats.wrap('redchi2s', redchi2s, [(0, 'dets')])
    
    if merge_stats:
        aman.wrap(hwpss_stats_name, hwpss_stats)
    if merge_model:
        aman.wrap(hwpss_model_name, fitsig_tod, [(0, 'dets'), (1, 'samps')])
    return hwpss_stats


def get_binned_hwpss(aman, signal=None, hwp_angle=None,
                     bins=360, flags=None, 
                     apodize_edges=True, apodize_edges_samps=1600,
                     apodize_flags=True, apodize_flags_samps=200):
    """
    Bin time-ordered data by the HWP angle and return the binned signal and its standard deviation.

    Parameters
    ----------
    aman : TOD
        The Axismanager object to be binned.
    signal : str, optional
        The name of the signal to be binned. Defaults to aman.signal if not specified.
    hwp_angle : str, optional
        The name of the timestream of hwp_angle. Defaults to aman.hwp_angle if not specified.
    bins : int, optional
        The number of HWP angle bins to use. Default is 360.
    flags : str or RangesMatrix or Ranges. optional
        Flag indicating whether to exclude flagged samples when binning the signal.
        If provided by a string, `aman.flags.get(flags)` is used for the flags.
        Default is no mask applied.
    apodize_edges : bool, optional
        If True, applies an apodization window to the edges of the signal. Defaults to True.
    apodize_edges_samps : int, optional
        The number of samples over which to apply the edge apodization window. Defaults to 1600.
    apodize_flags : bool, optional
        If True, applies an apodization window based on the flags. Defaults to True.
    apodize_flags_samps : int, optional
        The number of samples over which to apply the flags apodization window. Defaults to 200.

    Returns
    -------
    aman_proc:
        The AxisManager object which contains
        * center of each bin of hwp_angle
        * binned hwp synchrounous signal
        * estimated sigma of binned signal
    """
    if signal is None:
        signal = aman.signal
    if hwp_angle is None:
        hwp_angle = aman['hwp_angle']

    if apodize_edges:
        weight_for_signal = apodize.get_apodize_window_for_ends(aman, apodize_samps=apodize_edges_samps)
        if isinstance(flags, str):
            flags = aman.flags.get(flags)
        if (flags is not None) and apodize_flags:
            flags_mask = flags.mask()
            if flags_mask.ndim == 1:
                flag_is_1d = True
            else:
                all_columns_same = np.all(np.all(flags_mask == flags_mask[0, :], axis=0))
                if all_columns_same:
                    flag_is_1d = True
                    flags_mask = flags_mask[0]
                else:
                    flag_is_1d = False
            if flag_is_1d:
                weight_for_signal = weight_for_signal * apodize.get_apodize_window_from_flags(aman, 
                                                                                              flags=flags,
                                                                                              apodize_samps=apodize_flags_samps)
            else:
                weight_for_signal = weight_for_signal[np.newaxis, :] * apodize.get_apodize_window_from_flags(aman, 
                                                                                                             flags=flags, 
                                                                                                             apodize_samps=apodize_flags_samps)
    else:
        if (flags is not None) and apodize_flags:
            weight_for_signal = apodize.get_apodize_window_from_flags(aman, flags=flags, apodize_samps=apodize_flags_samps)
        else:
            weight_for_signal = None
    
    binning_dict = tod_ops.bin_signal(aman, bin_by=hwp_angle, range=[0, 2*np.pi],
                              bins=bins, signal=signal, flags=flags, weight_for_signal=weight_for_signal)
    
    bin_centers = binning_dict['bin_centers']
    bin_counts = binning_dict['bin_counts']
    binned_hwpss = binning_dict['binned_signal']
    binned_hwpss_sigma = binning_dict['binned_signal_sigma']
    
    # use median of sigma of each bin as uniform sigma for a detector
    hwpss_sigma = np.nanmedian(binned_hwpss_sigma, axis=-1)
    
    return bin_centers, bin_counts, binned_hwpss, hwpss_sigma


def hwpss_linreg(x, ys, yerrs, modes):
    """
    Performs a linear regression of the input data ys as a function of x, using a set of sine and cosine
    basis functions defined by the input modes. Returns the fitted signal, the coefficients of the
    basis functions, their covariance matrix, and the reduced chi-square.

    Parameters
    -----------
    x : numpy.ndarray
        The independent variable values of the data points to fit.
    ys : numpy.ndarray
        The dependent variable values of the data points to fit.
    yerrs : numpy.ndarray
        The error estimates of the dependent variable values.
    modes : list of int
        The frequencies of the sine and cosine basis functions to use.

    Returns
    -------
    fitsig : numpy.ndarray
        The fitted signal, obtained by evaluating the model with the optimal coefficients.
    coeffs : numpy.ndarray
        The coefficients of the sine and cosine basis functions that best fit the data.
    covars : numpy.ndarray
        The covariance matrix of the coefficients, estimated from the data errors.
    redchi2s : numpy.ndarray
        The reduced chi-square statistic of the fit, computed for each data point.
    """
    m = np.isnan(ys[0][:])
    xn = np.copy(x)
    x = x[~m]
    ys = ys[:, ~m]

    vects = np.zeros([2*len(modes), x.shape[0]], dtype='float32')
    for i, mode in enumerate(modes):
        vects[2*i, :] = np.sin(mode*x)
        vects[2*i+1, :] = np.cos(mode*x)

    I = np.linalg.inv(np.tensordot(vects, vects, (1, 1)))
    coeffs = np.matmul(ys, vects.T)
    coeffs = np.dot(I, coeffs.T).T
    fitsig = harms_func(x, modes, coeffs)

    # covariance of coefficients
    covars = np.zeros((ys.shape[0], 2*len(modes), 2*len(modes)))
    for det_idx in range(ys.shape[0]):
        covars[det_idx, :, :] = I * yerrs[det_idx]**2

    # reduced chi-square
    redchi2s = np.sum(
        ((ys - fitsig)/yerrs[:, np.newaxis])**2, axis=-1) / (x.shape[0] - 2*len(modes))

    fitsig = harms_func(xn, modes, coeffs)
    return fitsig, coeffs, covars, redchi2s


def harms_func(x, modes, coeffs):
    """
    calculates the harmonics function given the input values, modes and coefficients.

    Parameters
    ----------
    x (numpy.ndarray): Input values
    modes (list): List of modes to be used in the harmonics function
    coeffs (numpy.ndarray): Coefficients of the harmonics function

    Returns
    -------
    numpy.ndarray: The calculated harmonics function.
    """
    vects = np.zeros([2*len(modes), x.shape[0]], dtype=coeffs.dtype if coeffs is not None else 'float32')
    for i, mode in enumerate(modes):
        vects[2*i, :] = np.sin(mode*x)
        vects[2*i+1, :] = np.cos(mode*x)

    if coeffs is None:
        return vects
    else:
        harmonics = np.matmul(coeffs, vects, casting='no')
        return harmonics


def guess_hwpss_params(x, ys, modes):
    """
    Compute initial guess for the coefficients of a harmonics-based fit to data.

    Parameters
    ----------
    x : array-like of shape (nsamps,)
    ys : array-like of shape (ndets, nsamps)
    modes : array-like of shape (nmodes,)
        List of modes to use in the fit.

    Returns
    -------
    Params_init : ndarray of shape (m, 2*p)
        Initial guess for the coefficients of a harmonics-based fit to the data.
    """
    m = np.isnan(ys[0][:])
    x = x[~m]
    ys = ys[:, ~m]

    vects = harms_func(x, modes, coeffs=None)
    for i, mode in enumerate(modes):
        vects[2*i, :] = np.sin(mode*x)
        vects[2*i+1, :] = np.cos(mode*x)
    params_init = 2 * np.matmul(ys, vects.T) / x.shape[0]
    return params_init

def wrapper_harms_func(x, modes, *args):
    """
    A wrapper function for the harmonics function to be used for fitting data using Scipy's curve-fitting algorithm.
    Parameters
    ----------
    x : array-like
        The x-values of the data points to be fitted.
    modes : array-like
        An array of integers representing the modes of the harmonics function.
    *args : tuple
        A tuple of arguments. The first argument should be an array of coefficients used to calculate the harmonics function.
    Returns
    -------
    y : array-like
        An array of the same length as x representing the values of the harmonics function evaluated at x using the given 
        modes and coefficients.
    """
    coeffs = np.array(args[0])
    return harms_func(x, modes, coeffs)

def hwpss_curvefit(x, ys, yerrs, modes, params_init=None):
    """
    Fit harmonics to input data using scipy's curve_fit method.

    Parameters
    ----------
    x : array_like
        1-D array of x values.
    ys : array_like
        2-D array of y values for each detector.
    yerrs : array_like
        1-D array of the standard deviation of the y values for each detector.
    modes : array_like
        1-D array of mode numbers to be fitted.
    params_init : array_like, optional
        2-D array of initial parameter values for each detector. Default is None.

    Returns
    -------
    fitsig : ndarray
        2-D array of the fitted values for each detector.
    coeffs : ndarray
        2-D array of the fitted coefficients for each detector.
    covars : ndarray
        3-D array of the covariance matrix of the fitted coefficients for each detector.
    redchi2s : ndarray
        1-D array of the reduced chi-square values for each detector.

    Notes
    -----
    This function fits a set of harmonic functions to the input data using scipy's curve_fit method.
    The `modes` parameter specifies the mode numbers to be fitted.
    The `params_init` parameter can be used to provide initial guesses for the fit parameters.
    """
    N_dets = ys.shape[0]
    N_samps = ys.shape[-1]
    N_modes = len(modes)
    
    # Handle binned data w/ 0 counts in a bin.
    m = np.isnan(ys[0][:])
    xn = np.copy(x)
    x = x[~m]
    ys = ys[:, ~m]
  
    if params_init is None:
        params_init = np.zeros((N_dets, 2*N_modes))

    coeffs = np.zeros((N_dets, 2*len(modes)))
    covars = np.zeros((N_dets, 2*len(modes), 2*len(modes)))
    redchi2s = np.zeros(N_dets)
    fitsig = np.zeros((N_dets, N_samps))

    for det_idx in range(N_dets):
        p0 = params_init[det_idx]
        coeff, covar = curve_fit(lambda x, *p0: wrapper_harms_func(x, modes, p0),
                                 x, ys[det_idx], p0=p0, sigma=yerrs[det_idx] *
                                 np.ones_like(ys[det_idx]),
                                 absolute_sigma=True)

        coeffs[det_idx, :] = coeff
        covars[det_idx, :] = covar

        yfit = harms_func(x, modes, coeff)
        fitsig[det_idx, :] = harms_func(xn, modes, coeff)
        redchi2s[det_idx] = np.sum(
            ((ys[det_idx] - yfit) / yerrs[det_idx])**2) / (x.shape[0] - 2*len(modes))

    return fitsig, coeffs, covars, redchi2s


def estimate_sigma_tod(signal, hwp_angle):
    """
    Estimate the noise level of a signal in a time-ordered data (TOD) using a half-wave plate (HWP) modulation.

    Parameters
    ----------
    signal : ndarray
        A 2D numpy array of shape (n_dets, n_samps) containing the TOD of each detector.
    hwp_angle : ndarray
        A 1D numpy array containing the HWP angles in degrees.

    Returns
    -------
    hwpss_sigma_tod : ndarray
        A 1D numpy array containing the estimated noise level for each detector.

    Notes
    -----
    This function computes the mean of the signal in each period of HWP rotation and multiplies it
    by the square root of the number of samples in that period. The standard deviation of the
    resulting values for all periods is then computed and returned as the estimated sigma of each data point.
    """
    hwp_zeros_idxes = np.where(np.abs(np.diff(hwp_angle)) > 5)[0][:] + 1
    hwpss_sigma_tod = np.zeros((signal.shape[0], hwp_zeros_idxes.shape[0] - 1))

    for i, (init_idx, end_idx) in enumerate(zip(hwp_zeros_idxes[:-1], hwp_zeros_idxes[1:])):
        hwpss_sigma_tod[:, i] = np.mean(
            signal[:, init_idx:end_idx], axis=-1) * np.sqrt(end_idx - init_idx)
    hwpss_sigma_tod = np.std(hwpss_sigma_tod, axis=-1)
    return hwpss_sigma_tod


def subtract_hwpss(aman, signal='signal', hwpss_template_name='hwpss_model',
                   subtract_name='hwpss_remove', in_place=False, remove_template=True):
    """
    Subtract the half-wave plate synchronous signal (HWPSS) template from the
    signal in the given axis manager.

    Parameters
    ----------
    aman : AxisManager
        The axis manager containing the signal and the HWPSS template.
    signal : str, optional
        The name of the field in the axis manager containing the signal to be processed.
        Defaults to 'signal'.
    hwpss_template_name : str, optional
        The name of the field in the axis manager containing the HWPSS template.
        Defaults to 'hwpss_model'.
    subtract_name : str, optional
        The name of the field in the axis manager that will store the HWPSS-subtracted signal.
        Only used if in_place is False. Defaults to 'hwpss_remove'.
    in_place : bool, optional
        If True, the subtraction is done in place, modifying the original signal in the axis manager.
        If False, the result is stored in a new field specified by subtract_name. Defaults to False.
    remove_template : bool, optional
        If True, the HWPSS template field is removed from the axis manager after subtraction.
        Defaults to True.

    Returns
    -------
    None
    """
    if signal is None:
        signal_name = 'signal'
        signal = aman[signal_name]
    elif isinstance(signal, str):
        signal_name = signal
        signal = aman[signal_name]
    elif isinstance(signal, np.ndarray):
        if np.shape(signal) != (aman.dets.count, aman.samps.count):
            raise ValueError("When passing signal as ndarray shape must match (n_dets x n_samps).")
        signal_name = None
    else:
        raise TypeError("Signal must be None, str, or ndarray")

    if in_place:
        if signal_name is None:
            signal -= aman[hwpss_template_name].astype(signal.dtype)
        else:
            aman[signal_name] -= aman[hwpss_template_name].astype(aman[signal_name].dtype)
    else:
        if subtract_name in aman._fields:
            aman[subtract_name] = np.subtract(signal, aman[hwpss_template_name], dtype='float32')
        else:
            aman.wrap(subtract_name, np.subtract(signal,
                      aman[hwpss_template_name], dtype='float32'),
                      [(0, 'dets'), (1, 'samps')])
    
    if remove_template:
        aman.move(hwpss_template_name, None)

def get_hwp_freq(timestamps, hwp_angle):
    """
    Calculate the frequency of HWP rotation.

    Parameters:
    timestamps (array-like): An array of timestamps.
    hwp_angle (array-like): An array of HWP angles in radian

    Returns:
    float: The frequency of the HWP rotation in Hz.
    """
    hwp_freq = (np.sum(np.abs(np.diff(np.unwrap(hwp_angle)))) /
            (timestamps[-1] - timestamps[0])) / (2 * np.pi)
    return hwp_freq

def demod_tod(aman, signal=None, demod_mode=4,
              bpf_cfg=None, lpf_cfg=None, wrap=True):
    """
    Demodulate TOD based on HWP angle

    Parameters
    ----------
    aman : AxisManager
        The AxisManager object
    signal : str, optional
        Axis name of the signal to demodulate in aman. Default is 'signal'.
    demod_mode : int, optional
        Demodulation mode. Default is 4 (i.e. 4th harmonic of HWP).
    bpf_cfg : dict
        Configuration for Band-pass filter applied to the TOD data before demodulation.
        If not specified, a sine-squared bandwidth filter of
        (demod_mode * HWP speed) +/- 0.95*(HWP speed) is used with transition width 0.1.
        Example) bpf_cfg = {'type': 'sine2', 'center': 8.0, 'width': 3.8, 'trans_width': 0.1}
        or bpf_cfg = {'type': 'sine2', 'center': '4*f_HWP', 'width': '1.9*f_HWP', 'trans_width': 0.1}
        See filters.get_bpf for details.
    lpf_cfg : dict
        Configuration for Low-pass filter applied to the demodulated TOD data. If not specified,
        a sine-squared filter with a cutoff frequency of 0.95*(HWP speed) and transition width
        0.1 is used.
        Example) lpf_cfg = {'type': 'sine2', 'cutoff': 1.9, 'trans_width': 0.1}
        or lpf_cfg = {'type': 'sine2', 'cutoff': '0.95*f_HWP', 'trans_width': 0.1}
        See filters.get_lpf for details.
    wrap : bool, optional
        If True, the demodulated signal is wrapped and stored in the input aman container.
        If False, the demodulated signal is returned.

    Returns
    -------
    None
        The demodulated TOD data is added to the input `aman` container as new signals:
        'dsT' for the original signal filtered with `lpf`, 'demodQ' for the demodulated
        signal real component filtered with `lpf` and multiplied by 2, and 'demodU' for
        the demodulated signal imaginary component filtered with `lpf` and multiplied by 2.

    """
    if signal is None:
        #signal_name variable to be deleted when tod_ops.fourier_filter is updated
        signal_name = 'signal'
        signal = aman[signal_name]
    elif isinstance(signal, str):
        signal_name = signal
        signal = aman[signal_name]
    elif isinstance(signal, np.ndarray):
        raise TypeError("Currently ndarray not supported, need update to tod_ops.fourier_filter module to remove signal_name argument.")
    else:
        raise TypeError("Signal must be None, str, or ndarray")
    
    # HWP speed in Hz
    speed = get_hwp_freq(timestamps=aman.timestamps, hwp_angle=aman.hwp_angle)
    
    if bpf_cfg is None:
        bpf_center = demod_mode * speed
        bpf_width = speed * 2. * 0.95
        bpf_cfg = {'type': 'sine2',
                   'center': bpf_center,
                   'width': bpf_width,
                   'trans_width': 0.1}
    
    for k, v in bpf_cfg.items():
        if isinstance(v, str):
            if k in ['center', 'width']:
                bpf_cfg[k] = speed*float(v.split('*')[0])

    bpf = filters.get_bpf(bpf_cfg)
    
    if lpf_cfg is None:
        lpf_cutoff = speed * 0.95
        lpf_cfg = {'type': 'sine2',
                   'cutoff': lpf_cutoff,
                   'trans_width': 0.1}
    
    for k, v in lpf_cfg.items():
        if isinstance(v, str):
            if k == 'cutoff':
                lpf_cfg[k] = speed*float(v.split('*')[0])

    lpf = filters.get_lpf(lpf_cfg)

    # Create a RFFT object to reuse
    n = fft_ops.find_superior_integer(aman.samps.count)
    rfft = fft_ops.RFFTObj.for_shape(aman.dets.count, n, 'BOTH')

    phasor = np.empty_like(aman.hwp_angle, dtype=np.promote_types(signal.dtype, np.complex64))
    np.exp((demod_mode * 1j * aman.hwp_angle), out=phasor)
    demod = tod_ops.fourier_filter(aman, bpf, detrend=None,
                                   signal_name=signal_name, rfft=rfft) * 2. * phasor

    # Filter the demodulated signal
    demod_aman = core.AxisManager(aman.dets, aman.samps)
    demod_aman.wrap("timestamps", aman.timestamps, axis_map=[(0, 'samps')])
    demod_aman.wrap("dsT", aman[signal_name].copy(), axis_map=[(0, 'dets'), (1, 'samps')])
    demod_aman["dsT"][:] = tod_ops.fourier_filter(demod_aman, lpf, signal_name='dsT', detrend=None, rfft=rfft)
    demod_aman.wrap("demodQ", demod.real, axis_map=[(0, 'dets'), (1, 'samps')])
    demod_aman["demodQ"][:] = tod_ops.fourier_filter(demod_aman, lpf, signal_name="demodQ", detrend=None, rfft=rfft)
    demod_aman.wrap("demodU", demod.imag, axis_map=[(0, 'dets'), (1, 'samps')])
    demod_aman["demodU"][:] = tod_ops.fourier_filter(demod_aman, lpf, signal_name="demodU", detrend=None, rfft=rfft)

    # Destroy the RFFT object
    del rfft

    # Either wrap or return the demodulated signal
    if wrap:
        for fld in ['dsT', 'demodQ', 'demodU']:
            aman.wrap(fld, demod_aman[fld], axis_map=[(0, 'dets'), (1, 'samps')])
        del demod_aman
    else:
        return demod_aman['dsT'], demod_aman['demodQ'], demod_aman['demodU']


def tau_model(fhwp, tau, AQ, AU, mode):
    return (AQ + 1j*AU) / (1 - 1j* mode * 2 * np.pi * tau * fhwp )


def get_tau_hwp(
    aman,
    signal='signal',
    lpf_cfg=None,
    bpf_cfg=None,
    width=1000,
    apodize_samps=2000,
    trim_samps=2000,
    min_fhwp=1.,
    max_fhwp=2.,
    demod_mode=4,
    wn=None,
    flags=None,
    full_output=False,
    merge=False,
    name='tau_hwp'
):
    """
    Analyze observation with hwp spinning up or spinning down and
    compute the timeconstant of detectors.
    This demoulate tod and estimate timeconstant from hwp rotation
    speed depdndence of half-wave plate synchronous signal.

    Parameters
    ----------
    aman : AxisManager
        AxisManager object containing the TOD data.
    signal: std optional
        Name of signal to process. Default is `signal`.
    lpf_cfg: dict optional
        Configuration for Low-pass filter applied before demodulation.
    bpf_cfg: dict optional
        Configuration for Band-pass filter applied before demodulation.
    width: int optional
        width of single section of TOD.
    apodize_samps: int optional
        Number of samples on tod ends to apodize.
    trim_samps: int optional
        Number of samples on tod ends to trim.
    min_fhwp: float optional
        Mininum rotation frequency of hwp to be used for calculation.
    max_fhwp: float optional
        Maximum rotation frequency of hwp to be used for calculation.
    demod_mode: int, optional
        Demodulation mode. Default is 4.
    flags: str optional
        Name of flags in `aman.flags` to use for fitting.
    wn: str or None
        Precomputed white noise level of signal to be used for weights of
        fitting. If none, the fitting weights will be 1.
    full_output: bool optional
        Whether to output all the statistics used for fitting.
    merge: bool optional
        Whether to merge calculated statistics to `aman`. Default is False.
    name: str optional
        Name under which to wrap the calculated statistics.
        Default is 'tau_hwp'.

    Returns
    -------
    result : AxisManager
        An AxisManager containing time constants, their errors, and reduced
        chi-squared statistics.
    """

    if lpf_cfg is None:
        lpf_cfg = {'type': 'sine2',
                   'cutoff': min_fhwp * 0.95,
                   'trans_width': 0.1}
    if bpf_cfg is None:
        # no band pass filter
        bpf_cfg = {'type': 'sine2',
                   'center': 0,
                   'width': 200,
                   'trans_width': 0.1}

    hwp_rate = np.gradient(np.unwrap(aman.hwp_angle) * 200 / 2 / np.pi)
    if 'hwp_rate' not in aman:
        aman.wrap('hwp_rate', hwp_rate, [(0, 'samps')])

    if wn is None:
        # Calculate white noise from section with low hwp speed.
        idx = np.where(abs(hwp_rate) < .5)[0]
        s = aman.samps.offset + idx[0]
        e = aman.samps.offset + idx[-1]
        aman_short = aman.restrict('samps', (s, e), in_place=False)
        freqs, Pxx, nseg = fft_ops.calc_psd(aman_short,
                                            signal=aman_short.get(signal),
                                            noverlap=0,
                                            full_output=True)
        wn = fft_ops.calc_wn(aman_short, pxx=Pxx, freqs=freqs, nseg=nseg)

    tod_ops.detrend_tod(aman, signal_name=signal, method="median")
    tod_ops.apodize.apodize_cosine(aman, signal_name=signal,
                                   apodize_samps=apodize_samps)
    demod_tod(aman, signal=signal, demod_mode=demod_mode,
              lpf_cfg=lpf_cfg, bpf_cfg=bpf_cfg, wrap=True)

    idx = np.where((min_fhwp < abs(hwp_rate)) & (abs(hwp_rate) < max_fhwp))[0]
    s = idx[0] if trim_samps < idx[0] else trim_samps
    e = idx[-1] if trim_samps < aman.samps.count - idx[-1] \
        else aman.samps.count - trim_samps
    s += aman.samps.offset
    e += aman.samps.offset

    aman.restrict('samps', (s, e), in_place=True)

    nsections = int(aman.samps.count / width)
    timestamps = np.zeros(nsections)
    hwp_rate = np.zeros(nsections)
    demodQ = np.zeros((nsections, aman.dets.count))
    demodU = np.zeros((nsections, aman.dets.count))
    weights = np.zeros((nsections, aman.dets.count))
    mask = np.ones((nsections, aman.dets.count), dtype=bool)
    if isinstance(flags, str) and (flags in aman.flags):
        flags = ~aman.flags[flags].mask()
        if flags.ndim == 1:
            for i in range(nsections):
                msk = np.zeros(aman.samps.count, dtype=bool)
                msk[i*width:(i+1)*width] = flags[i*width:(i+1)*width]
                timestamps[i] = np.median(aman.timestamps[msk])
                hwp_rate[i] = np.median(aman.hwp_rate[msk])
                demodQ[i] = np.median(aman.demodQ[:, msk], axis=1)
                demodU[i] = np.median(aman.demodU[:, msk], axis=1)
                count = np.sum(msk)
                weights[i] = np.sqrt(count/200)/wn
                mask[i, :] = False if count == 0 else True
        elif flags.ndim == 2:
            for i in range(nsections):
                timestamps[i] = np.median(aman.timestamps[i*width:(i+1)*width])
                hwp_rate[i] = np.median(aman.hwp_rate[i*width:(i+1)*width])
                for j in range(aman.dets.count):
                    msk = np.zeros(aman.samps.count, dtype=bool)
                    msk[i*width:(i+1)*width] = flags[j, i*width:(i+1)*width]
                    demodQ[i, j] = np.median(aman.demodQ[j, msk])
                    demodU[i, j] = np.median(aman.demodU[j, msk])
                    count = np.sum(msk)
                    weights[i, j] = np.sqrt(count/200)/wn[j]
                    mask[i, j] = False if count == 0 else True
    else:
        for i in range(nsections):
            timestamps[i] = np.median(aman.timestamps[i*width:(i+1)*width])
            hwp_rate[i] = np.median(aman.hwp_rate[i*width:(i+1)*width])
            demodQ[i] = np.median(aman.demodQ[:, i*width:(i+1)*width], axis=1)
            demodU[i] = np.median(aman.demodU[:, i*width:(i+1)*width], axis=1)
            weights[i] = np.sqrt(width/200)/wn

    logger.debug('Fit time constant')
    AQ = np.full(aman.dets.count, np.nan)
    AQ_error = np.full(aman.dets.count, np.nan)
    AU = np.full(aman.dets.count, np.nan)
    AU_error = np.full(aman.dets.count, np.nan)
    tau = np.full(aman.dets.count, np.nan)
    tau_error = np.full(aman.dets.count, np.nan)
    redchi2s = np.full(aman.dets.count, np.nan)
    for i in range(aman.dets.count):
        try:
            model = LmfitModel(tau_model, independent_vars=['fhwp'])
            model.set_param_hint('mode', vary=False)
            params = model.make_params(
                tau=1e-3,
                AQ=np.median(demodQ[:, i][mask[:, i]]),
                AU=np.median(demodU[:, i][mask[:, i]]),
                mode=demod_mode,
            )
            fit = model.fit(
                data=demodQ[:, i][mask[:, i]] + 1j * demodU[:, i][mask[:, i]],
                params=params,
                fhwp=hwp_rate[mask[:, i]],
                weights=weights[:, i][mask[:, i]]
            )
            AQ[i] = fit.params['AQ'].value
            AQ_error[i] = fit.params['AQ'].stderr
            AU[i] = fit.params['AU'].value
            AU_error[i] = fit.params['AU'].stderr
            tau[i] = fit.params['tau'].value
            tau_error[i] = fit.params['tau'].stderr
            redchi2s[i] = fit.redchi
        except Exception:
            logger.debug(f'Failed to fit {aman.dets.vals[i]}')

    if full_output:
        sections = core.OffsetAxis('sections', count=nsections,
                                   offset=0)
        result = core.AxisManager(aman.dets, sections)
        result.wrap('timestamps', timestamps, axis_map=[(0, 'sections')])
        result.wrap('hwp_rate', hwp_rate, axis_map=[(0, 'sections')])
        result.wrap('demodQ', np.transpose(demodQ),
                    axis_map=[(0, 'dets'), (1, 'sections')])
        result.wrap('demodU', np.transpose(demodU),
                    axis_map=[(0, 'dets'), (1, 'sections')])
        result.wrap('weights', np.transpose(weights),
                    axis_map=[(0, 'dets'), (1, 'sections')])
        result.wrap('mask', np.transpose(mask),
                    axis_map=[(0, 'dets'), (1, 'sections')])
    else:
        result = core.AxisManager(aman.dets)
    result.wrap('AQ', AQ, axis_map=[(0, 'dets')])
    result.wrap('AQ_error', AQ_error, axis_map=[(0, 'dets')])
    result.wrap('AU', AU, axis_map=[(0, 'dets')])
    result.wrap('AU_error', AU_error, axis_map=[(0, 'dets')])
    result.wrap('tau_hwp', tau, axis_map=[(0, 'dets')])
    result.wrap('tau_hwp_error', tau_error, axis_map=[(0, 'dets')])
    result.wrap('redchi2s', redchi2s, axis_map=[(0, 'dets')])

    if merge:
        aman.wrap(name, result)

    return result



import unittest
from sotodlib.core import metadata

import os
import time
import numpy as np

from ._helpers import mpi_multi

# Global Announcement: I know, but I hate slow tests.
example = None


@unittest.skipIf(mpi_multi(), "Running with multiple MPI processes")
class TestDetDb(unittest.TestCase):
    def setUp(self):
        global example
        if example is None:
            print('Creating example database...')
            example = metadata.get_example('DetDb')

    def test_smoke(self):
        """Basic functionality."""
        db = example.copy()

        print('Test 1: how many dets have array_code=HF1?')
        X = db.dets(props={'base.array_code': 'HF1'})
        print('  Answer: %i' % len(X))
        print('  The first few are:')
        for x in X[:5]:
            print('    ' + str(x))
        print()

        print('Test 2: Get (array, wafer) for a bunch of dets.')
        u2 = db.props(X, props=['base.array_code', 'wafer_code'])
        pairs = list(u2.distinct())
        print('  Distinct pairs:')
        for p in pairs:
            print('    ' + str(p))
        print()
        assert(len(pairs) == 3)

    def test_resultset(self):
        """Test that ResultSet objects have required behaviors.

        """
        db0 = example.copy()
        dets = db0.dets()
        props = db0.props(props=['base.array_code', 'base.wafer_code'])
        combos = props.distinct()

        assert isinstance(dets, metadata.ResultSet)
        assert isinstance(props, metadata.ResultSet)
        assert isinstance(combos, metadata.ResultSet)

        assert isinstance(combos[0], dict)
        assert isinstance(list(combos)[0], dict)

        # Test distinct coverage.
        n0 = len(db0.dets(props=combos))
        n1 = 0
        for c in combos:
            subd = db0.dets(props=c)
            n1 += len(subd)
        assert(n0 == n1)

        # Check operators...
        assert isinstance(combos[:2] + combos[2:], metadata.ResultSet)
        with self.assertRaises(TypeError):
            combos + [1, 2, 3]
        with self.assertRaises(ValueError):
            combos + dets

        # Check indexing
        # ... with int
        self.assertIsInstance(dets[1], dict)
        self.assertIsInstance(dets[int(1)], dict)
        self.assertIsInstance(dets[np.arange(3)[1]], dict)

        # ... with slice
        dets_subset = dets[1:3]
        self.assertIsInstance(dets_subset, metadata.ResultSet)
        self.assertEqual(len(dets_subset), 2)

        # ... Check that string index returns field
        names = dets['name']
        self.assertIsInstance(names, np.ndarray)
        self.assertEqual(len(names), len(dets))

    def test_io(self):
        """Check to_file and from_file."""

        db0 = example.copy()
        dump_list = [(f'test.sqlite', None),
                     (f'test.txt', 'dump'),
                     (f'test.gz', None)]
        # Save.
        for fn, fmt in dump_list:
            print(f'Writing {fn}')
            db0.to_file(fn, fmt=fmt)
            print('  -- output has size {}'.format(os.path.getsize(fn)))
            t0 = time.time()
            db1 = metadata.DetDb.from_file(fn, fmt=fmt)
            dt = time.time() - t0
            print('  -- read-back {} rows in {} seconds.'.format(
                len(db1.dets()), dt))
            print('  -- removing.')
            os.remove(fn)


if __name__ == '__main__':
    unittest.main()



import sqlite3
import gzip
import os

from .resultset import ResultSet
from . import common


class SchemaError(Exception):
    """
    This is raised in cases where the code detects a schema violation,
    such as tables not having the required named columns.
    """
    pass


class IntervalError(Exception):
    """
    This is raised in cases where the code detects that time intervals
    in a property table are of negative size or overlap with other
    intervals for the same det_id.
    """
    pass


TABLE_DEFS = {
    'dets': [
        "`id` integer primary key autoincrement",
        "`name` varchar(256) unique",
        ],
}
SPECIAL_COLS = ['det_id', 'time0', 'time1']


class DetDb(object):
    """
    Detector database.  The database stores data about a set of
    detectors.

    The ``dets`` table lists all valid detectors, associating a
    (time-invariant) name to each ``id``.

    The other tables in the database are user configurable "property
    tables" that must obey certain rules:

    1. They have at least the following columns:

       - ``det_id`` integer
       - ``time0`` integer (unix timestamp)
       - ``time1`` integer (unix timestamp)

    2. The values time0 and time1 define an interval ``[time0,time1)``
       over which the data in the row is valid.  Every row shall
       respect the constraint that ``time0 <= time1``.

    3. No two rows in a property table shall have the same ``det_id``
       and overlapping time intervals.  Note that since the intervals
       are half-open, the intervals [t0, t1) and [t1, t2) do not
       overlap.
    """

    #: A time-range that is meant to signify "all reasonable times";
    #: in this case it spans from years 1970 - 2096.
    ALWAYS = (0., 4e9)

    #: Column definitions (a list of strings) that must appear in all
    #: Property Tables.
    TABLE_TEMPLATE = [
        "`det_id` integer",
        "`time0` integer",
        "`time1` integer",
    ]

    def __init__(self, map_file=None, init_db=True):
        """Instantiate a DetDb.  If map_file is provided, the database will
        be connected to the indicated sqlite file on disk, and any
        changes made to this object be written back to the file.

        """
        if isinstance(map_file, sqlite3.Connection):
            self.conn = map_file
        else:
            if map_file is None:
                map_file = ':memory:'
            self.conn = sqlite3.connect(map_file)
        self.conn.row_factory = sqlite3.Row  # access columns by name

        if init_db:
            # Create dets table if not found.
            c = self.conn.cursor()
            c.execute("SELECT name FROM sqlite_master "
                      "WHERE type='table' and name not like 'sqlite_%';")
            tables = [r[0] for r in c]
            if 'dets' not in tables:
                self.create_table('dets', TABLE_DEFS['dets'], raw=True)

    def __len__(self):
        return self.conn.execute('select count(id) from dets').fetchone()[0]

    def _get_property_tables(self):
        """Return a list of all property tables."""
        c = self.conn.cursor()
        c.execute("SELECT name FROM sqlite_master WHERE "
                  "type='table' and name not like 'sqlite_%' "
                  "and name != 'dets';")
        return [str(x[0]) for x in c]

    def _get_property_list(self, table_name):
        """Return a list of all property columns for the specified property
        table.

        """
        c = self.conn.cursor()
        c.execute(f'PRAGMA table_info({table_name})')
        return [r[1] for r in c if r[1] not in SPECIAL_COLS]

    def validate(self):
        """
        Checks that the database is following internal rules.
        Specifically we check that a ``dets`` table exists and has the
        necessary columns; then we check that all other tables do not
        have overlapping property intervals.  Raises SchemaError in
        the first case, IntervalError in the second.
        """
        c = self.conn.cursor()
        c.execute("SELECT name FROM sqlite_master WHERE type='table' "
                  "and name not like 'sqlite_%';")
        tables = [r[0] for r in c]
        if 'dets' not in tables:
            raise SchemaError("Database does not contain a `dets` table.")
        tables.remove('dets')
        for t in tables:
            try:
                c.execute("SELECT det_id,time0,time1 from `%s` "
                          "order by det_id,time0" % t)
            except sqlite3.OperationalError as e:
                raise SchemaError("Key columns not found in table `%s`" % t)
            last_id, last_t1 = None, None
            for r in c:
                _id, _t0, _t1 = r
                if (_t1 < _t0):
                    raise IntervalError(
                        "Negative size time interval for table %s, "
                        "det_id=%i" % (t, _id))
                if _id == last_id:
                    if _t0 < last_t1:
                        raise IntervalError(
                            "Overlapping interval for table %s, "
                            "det_id=%i" % (t, _id))
                    last_t1 = _t1
                else:
                    last_id, last_t1 = _id, _t1

    def create_table(self, table_name, column_defs, raw=False, commit=True):
        """Add a property table to the database.

        Args:
          table_name (str): The name of the new table.
          column_defs (list): A list of sqlite column definition
            strings.
          raw (bool): See below.
          commit (bool): Whether to commit the changes to the db.

        The special columns `det`, `time0` and `time1` will be
        pre-pended unless raw=True.  An example of column_defs is::

          column_defs=[
            "`x_pos` float",
            "`y_pos` float",
          ]

        """
        c = self.conn.cursor()
        pre_cols = self.TABLE_TEMPLATE
        if raw:
            pre_cols = []
        q = ('create table if not exists `%s` (' % table_name +
             ','.join(pre_cols + column_defs) + ')')
        c.execute(q)
        if commit:
            self.conn.commit()
        return self

    def copy(self, map_file=None, overwrite=False):
        """
        Duplicate the current database into a new database object, and
        return it.  If map_file is specified, the new database will be
        connected to that sqlite file on disk.  Note that a quick way
        of writing a Db to disk to call copy(map_file=...) and then
        simply discard the returned object.
        """
        if map_file is not None and os.path.exists(map_file):
            if overwrite:
                os.remove(map_file)
            else:
                raise RuntimeError("Output file %s exists (overwrite=True "
                                   "to overwrite)." % map_file)
        new_db = DetDb(map_file=map_file, init_db=False)
        script = ' '.join(self.conn.iterdump())
        new_db.conn.executescript(script)
        return new_db

    def to_file(self, filename, overwrite=True, fmt=None):
        """Write the present database to the indicated filename.

        Args:
          filename (str): the path to the output file.
          overwrite (bool): whether an existing file should be
            overwritten.
          fmt (str): 'sqlite', 'dump', or 'gz'.  Defaults to 'sqlite'
            unless the filename ends with '.gz', in which it is 'gz'.

        """
        if fmt is None:
            if filename.endswith('.gz'):
                fmt = 'gz'
            else:
                fmt = 'sqlite'
        if os.path.exists(filename) and not overwrite:
            raise RuntimeError(f'File {filename} exists; remove or pass '
                               'overwrite=True.')
        if fmt == 'sqlite':
            self.copy(map_file=filename, overwrite=overwrite)
        elif fmt == 'dump':
            with open(filename, 'w') as fout:
                for line in self.conn.iterdump():
                    fout.write(line)
        elif fmt == 'gz':
            with gzip.GzipFile(filename, 'wb') as fout:
                for line in self.conn.iterdump():
                    fout.write(line.encode('utf-8'))
        else:
            raise RuntimeError(f'Unknown format "{fmt}" requested.')

    @classmethod
    def from_file(cls, filename, fmt=None, force_new_db=True):
        """This method calls
            :func:`sotodlib.core.metadata.common.sqlite_from_file`
        """
        conn = common.sqlite_from_file(filename, fmt=fmt, 
                                       force_new_db=force_new_db)
        return cls(conn, init_db=False)
        

    def reduce(self, dets=None, time0=None, time1=None,
               inplace=False):
        """Discard information from the database unless it is "relevant".

        Args:
          dets (list or ResultSet): A list of detectors names that
            are relevant.  If this is a ResultSet, the 'name' column
            is used.  If None, all dets are relevant.
          time0 (int): If time1 is None, then a property's time range
            must contain this time for it to be considered relevant.
            If time1 is not None, see below.
          time1 (int): Along with time0, forms a time range that have
            non-zero intersection with the property's time range for
            the entry to be considered relevant.
          inplace (bool): Whether to act on the present object, or to
            return a modified copy.

        Returns the reduced data (which is self, if inplace is True).

        """
        if not inplace:
            return self.copy().reduce(dets, time0, time1, inplace=True)

        time_clause = '0'
        if time0 is not None:
            if time1 is None:
                time_clause = '(%i >= time0) or (time1 > %i)' % (time0, time0)
            else:
                assert(time1 >= time0)
                time_clause = '(%i <= time1) or (time0 < %i)' % (time0, time1)
        else:
            assert(time1 is None)

        c = self.conn.cursor()
        if dets is not None:
            # Convert to a list of names.
            if isinstance(dets, ResultSet):
                dets = dets['name']
            # Create a temporary table to list dets we're keeping.
            self.create_table('_keepers', ["`name` varchar"], raw=True)
            for d in dets:
                c.execute('insert into _keepers (name) values (?)', (d,))
            # Intersect dets against _keepers, discard _keepers.
            c.execute('delete from dets where not dets.name in '
                      '(select name from _keepers)')
            c.execute('drop table _keepers')
            det_clause = 'not det_id in (select id from dets)'
        else:
            det_clause = '0'

        # Remove orphaned rows from other tables.
        for t in self._get_property_tables():
            c.execute('delete from %s where %s or %s' %
                      (t, det_clause, time_clause))
        self.conn.commit()

        # Compact the db.
        c.execute('vacuum')
        self.conn.commit()

        return self

    # Construction.
    def get_id(self, name, commit=True, create=True):
        """Returns a detector's internal id.  If the detector isn't in the
        dets table yet, and create==True, then it is added.

        """
        c = self.conn.execute('select id from dets where name=?',
                              (name,))
        db_id = c.fetchone()
        if db_id is not None:
            return db_id[0]
        if not create:
            raise ValueError("Detector {} not in table and "
                             "create=False".format(name))
        c = self.conn.execute('insert into dets (name) values (?)',
                              (name,))
        db_id = c.lastrowid
        if commit:
            self.conn.commit()
        return db_id

    def add_props(self, table_, name_, time_range=None, commit=True, **kw):
        """Add property information for a detector.

        Args:
          table_ (str): The property table name.
          name_ (str): The detector name.
          time_range (pair of ints): The time range over which the
            property value is applicable.
          commit (bool): Whether or not to commit the db.

        All other keyword arguments are interpreted as data to write
        into the property table.

        """
        if time_range is None:
            time_range = self.ALWAYS
        row_id = self.get_id(name_, create=True, commit=False)
        keys, values = zip(*kw.items())
        key_string = ('det_id,time0,time1' +
                      (''.join([',`{}`'] * len(keys)).format(*keys)))
        val_string = '?,?,?' + ''.join([',?'] * len(keys))
        q = (f'insert into {table_} ({key_string}) '
             f'values ({val_string})')
        self.conn.execute(
            q, (row_id, time_range[0], time_range[1]) + tuple(values))
        if commit:
            self.conn.commit()

    # Forward lookup.
    def dets(self, timestamp=None, props={}):
        """
        Get a list of detectors matching the conditions listed in the
        "props" dict.  If timestamp is not provided, then time range
        restriction is not applied.

        Returns a list of detector names.
        """
        # Accumulate a query, and args.
        q = 'select dets.name as name from dets'
        args = []

        # Whatever we were given, convert it to a list of dicts.
        if isinstance(props, ResultSet):
            prop_sets = list(props.distinct())
        elif isinstance(props, dict):
            prop_sets = [props]
        else:
            prop_sets = props

        # Expand each match row into query and args.
        other_tables = []
        row_wheres = []
        for props in prop_sets:
            r = []
            for m, v in props.items():
                if '.' in m:
                    t, m = m.split('.', 1)
                else:
                    t, m = 'base', m
                if t not in other_tables:
                    other_tables.append(t)
                r.append((f'{t}.{m}=?', v))
            row_wheres.append(r)

        # Joins.
        for t in other_tables:
            q += ' join %s on %s.det_id=dets.id' % (t, t)
        # Accumulate restriction strings...
        restricts = []
        if timestamp is not None:
            time_clause = '%s.time0 <= ? and ? < %s.time1'
            for t in other_tables:
                restricts.append(time_clause % (t, t))
                args.extend([timestamp, timestamp])
        # Matching of each prop_set.
        prop_criteria = []
        for r in row_wheres:
            if len(r) == 0:
                prop_criteria.append('1')
                continue
            conds, vals = zip(*r)
            prop_criteria.append(' and '.join(conds))
            args.extend(vals)

        if len(prop_criteria) == 0:
            prop_criteria.append('0')
        restricts.append(' or '.join(['(' + pc + ')' for pc in prop_criteria]))

        # Apply restrictions...
        if (restricts):
            q += ' where ' + ' and '.join(restricts)
        q = q + ' group by id'
        c = self.conn.cursor()
        c.execute(q, tuple(args))
        return ResultSet.from_cursor(c)

    # Reverse lookup.
    def props(self, dets=None, timestamp=None, props=None,
              concise=False):
        """
        Get the value of the properties listed in props, for each detector
        identified in dets (a list of strings, or a ResultSet with a
        column called 'name').
        """
        # Create temporary table
        c = self.conn.cursor()
        c.execute('begin transaction')
        c.execute('drop table if exists _dets')
        c.execute('create temp table _dets (`name` varchar(32))')
        q = 'insert into _dets (name) values (?)'
        if dets is None:
            dets = self.dets()['name']
        if isinstance(dets, ResultSet):
            dets = dets['name']
        elif isinstance(dets, dict):
            # This is intended to handle a single row extracted from a
            # ResultSet.
            dets = [dets['name']]
        for a in dets:
            c.execute(q, (a,))
        c.execute('end transaction')

        # Expand props argument.
        if props is None:
            props = [t + '.' for t in self._get_property_tables()]

        props, props_ = [], props
        for p in props_:
            if p.endswith('.'):
                table_p = self._get_property_list(p[:-1])
                props.extend([p + _p for _p in table_p])
            else:
                props.append(p)

        # Now look stuff up in it.
        other_tables = []
        fields, keys = [], []
        for i, m in enumerate(props):
            if '.' in m:
                t, f = m.split('.', 1)
            else:
                t, f = 'base', m
            if t not in other_tables:
                other_tables.append(t)
            key = f'{t}.{f}'
            keys.append(key)
            fields.append(f'{key} as result{i}')
        q = ('select ' + ', '.join(fields) +
             ' from _dets join dets on _dets.name=dets.name ' +
             ' '.join(['join %s on %s.det_id=dets.id' % (m, m)
                       for m in other_tables]))
        c.execute(q)
        results = ResultSet.from_cursor(c, keys=keys)
        c.execute('drop table if exists _dets')
        results.strip(['base.'])
        return results

    def intersect(self, *specs, resolve=False):
        """Intersect the provided detector specs.  Each entry is either a list
        (or similar iterable) of detector names, or a dictionary
        specifying detector properties.

        If resolve=True, then the returned item is a list (rather
        than, possibly, a dict).

        """
        if len(specs) == 0:
            return []
        dicts = [s for s in specs if isinstance(s, dict)]
        others = [s for s in specs if not isinstance(s, dict)]
        # Reduce the dicts.
        req = {}
        for d in dicts:
            for k, v in d.items():
                if k in req:
                    if req[k] != v:
                        return []
                else:
                    req[k] = v
        if len(others) == 0:
            if resolve:
                return self.dets(props=req)['name']
            return req

        # Turn it into a list.
        req = self.dets(props=req)['name']
        keepers = set(req)
        for other in others:
            keepers.intersection_update(other)
        return [n for n in req if n in keepers]


def get_example():
    """Returns an example DetDb, mapped to RAM.  The two property tables
    are called "base" and "geometry".  This example is for
    demonstrating the code and interface and has no relation to any
    instrument's actual detector layout!

    """
    db = DetDb()

    TABLES = [
        ('base', [
            "`instrument` varchar(32)",
            "`camera` varchar(32)",
            "`array_code` varchar(16)",
            "`array_class` varchar(16)",
            "`wafer_code` varchar(32)",
            "`freq_code` varchar(16)",
            "`det_type` varchar(32)",
        ]),
        ('geometry', [
            "`wafer_x` float",
            "`wafer_y` float",
            "`wafer_pol` float",
        ]),
    ]

    for n, d in TABLES:
        print('Creating table %s' % n)
        db.create_table(n, d)

    tel_info = {'instrument': 'simonsobs',
                'camera': 'latr'}

    det_names = []
    for ar_type, bands, n_ar, n_wa, n_det in [
            ('LF', [27, 39], 1, 3, 37),
            ('MF', [93, 145], 4, 3, 432),
            ('HF', [225, 278], 2, 3, 542)
    ]:
        print('Creating %s-type arrays...' % ar_type)
        for fi, f in enumerate(bands):
            for ar in range(n_ar):
                for wa in range(n_wa):
                    iofs = (fi*n_ar*n_wa + n_wa*ar + wa)*n_det
                    info = {'freq_code': 'f%03i' % f,
                            'array_class': ar_type,
                            'array_code': '%s%i' % (ar_type, ar+1),
                            'wafer_code': 'W%i' % (wa+1)}
                    for i in range(iofs, iofs + n_det):
                        det_names.append('%s%i_%05i' % (ar_type, ar+1, i))
                        db.add_props('base', det_names[-1],
                                     det_type='bolo',
                                     **tel_info, **info, commit=False)

    print('Committing {} detectors...'.format(len(det_names)))
    db.conn.commit()

    # Organize these dets in a big square.  This is not the plan.
    n_row = int(len(det_names)**.5 + 1)
    for i in range(n_row):
        for j in range(n_row):
            d = i*n_row+j
            if d >= len(det_names):
                break
            x, y, ang = i * .02, j * .02, (i+j) % 12. * 15
            db.add_props('geometry', det_names[d], commit=False,
                         wafer_x=x, wafer_y=y, wafer_pol=ang)

    db.conn.commit()

    print('Checking the work...')
    db.validate()

    return db



# Copyright (c) 2018-2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Internal helper functions for unit tests.
"""

import os
import datetime
import shutil
import tempfile

import numpy as np
import astropy.units as u

from sotodlib import core, coords

try:
    import toast
    import sotodlib.toast as sotoast
    import sotodlib.toast.ops as so_ops
    import toast.schedule_sim_ground
    from toast.observation import default_values as defaults
    toast_available = True
except ImportError as e:
    toast_available = False

from sotodlib.sim_hardware import sim_nominal


def mpi_world():
    """Return the MPI world communicator or None.
    """
    comm = None
    procs = 1
    rank = 0
    if "MPI_DISABLE" not in os.environ:
        try:
            import mpi4py.MPI as MPI

            comm = MPI.COMM_WORLD
            procs = comm.size
            rank = comm.rank
        except Exception:
            pass
    return comm, procs, rank


def mpi_multi():
    """Return True if we have more than one MPI process in our environment.
    """
    comm, procs, rank = mpi_world()
    if procs > 1:
        return True
    else:
        return False


def create_outdir(subdir=None, mpicomm=None):
    """Create the top level output directory and per-test subdir.

    Args:
        subdir (str): the sub directory for this test.
        mpicomm (MPI.Comm):  Optional communicator.

    Returns:
        str: full path to the test subdir if specified, else the top dir.

    """
    rank = 0
    if mpicomm is not None:
        rank = mpicomm.rank
    retdir = None
    if rank == 0:
        pwd = os.path.abspath(".")
        testdir = os.path.join(pwd, "sotodlib_test_output")
        retdir = testdir
        if subdir is not None:
            retdir = os.path.join(testdir, subdir)
        if not os.path.isdir(testdir):
            os.mkdir(testdir)
        if not os.path.isdir(retdir):
            os.mkdir(retdir)
    if mpicomm is not None:
        retdir = mpicomm.bcast(retdir, root=0)
    return retdir


def close_data_and_comm(data):
    """Delete a toast data object AND the input Comm.

    Multiple toast.Data objects can be created from a single toast.Comm,
    and so the communicators are not freed when deleting the data object.
    However for unit tests we frequently use a helper function to produce
    a simulated dataset and then want to fully clean up that along with
    the communicators that were used.  This is especially true on CI
    services where repeatedly creating communicators without cleanup seem
    to cause sporadic deadlocks.  This is a convenience function which
    does that cleanup.

    Args:
        data (toast.Data):  The input data object

    Returns:
        None

    """
    cm = data.comm
    if cm.comm_world is not None:
        cm.comm_world.barrier()
    data.clear()
    del data
    cm.close()
    del cm


def observing_schedule(telescope, mpicomm=None, temp_dir=None):
    if not toast_available:
        print("toast unavailable- cannot build observing schedule")
        return
    schedule = None
    if mpicomm is None or mpicomm.rank == 0:
        tdir = temp_dir
        if tdir is None:
            tdir = tempfile.mkdtemp()

        sch_file = os.path.join(tdir, "schedule.txt")
        toast.schedule_sim_ground.run_scheduler(
            opts=[
                "--site-name",
                telescope.site.name,
                "--telescope",
                telescope.name,
                "--site-lon",
                "{}".format(telescope.site.earthloc.lon.to_value(u.degree)),
                "--site-lat",
                "{}".format(telescope.site.earthloc.lat.to_value(u.degree)),
                "--site-alt",
                "{}".format(telescope.site.earthloc.height.to_value(u.meter)),
                "--patch",
                "small_patch,1,40,-40,44,-44",
                "--el-min",
                "40",
                "--el-max",
                "60",
                "--start",
                "2025-01-01 00:00:00",
                "--stop",
                "2025-01-01 03:00:00",
                "--out",
                sch_file,
            ]
        )
        schedule = toast.schedule.GroundSchedule()
        schedule.read(sch_file)
        if temp_dir is None:
            shutil.rmtree(tdir)
    if mpicomm is not None:
        schedule = mpicomm.bcast(schedule, root=0)
    return schedule


def calibration_schedule(telescope):
    """Make a small observing schedule for calibration tasks.

    This is a fake schedule with no scanning.

    """
    if not toast_available:
        print("toast unavailable- cannot build calibration schedule")
        return

    scans = [
        toast.schedule.GroundScan(
            name="CAL",
            start=datetime.datetime(
                2025, 1, 1, hour=0, minute=0, second=0
            ),
            stop=datetime.datetime(
                2025, 1, 1, hour=0, minute=10, second=0
            ),
            boresight_angle=0 * u.degree,
            az_min=-1.0 * u.degree,
            az_max=1.0 * u.degree,
            el=60.0 * u.degree,
            scan_indx=0,
            subscan_indx=0,
        )
    ]

    schedule = toast.schedule.GroundSchedule(
        scans=scans,
        site_name=telescope.site.name,
        telescope_name=telescope.name,
        site_lat=telescope.site.earthloc.lat,
        site_lon=telescope.site.earthloc.lon,
        site_alt=telescope.site.earthloc.height,
    )

    return schedule


def create_comm(mpicomm):
    """Create a toast communicator.

    Use the specified MPI communicator to attempt to create 2 process groups.
    If less than 2 processes are used, create a single process group.

    Args:
        mpicomm (MPI.Comm): the MPI communicator (or None).

    Returns:
        toast.Comm: the 2-level toast communicator.

    """
    if not toast_available:
        raise RuntimeError("TOAST is not importable, cannot create a toast.Comm")
    toastcomm = None
    if mpicomm is None:
        toastcomm = toast.Comm(world=mpicomm)
    else:
        worldsize = mpicomm.size
        groupsize = 1
        if worldsize >= 2:
            groupsize = worldsize // 2
        toastcomm = toast.Comm(world=mpicomm, groupsize=groupsize)
    return toastcomm


def simulation_test_data(
    mpicomm,
    telescope_name="SAT4",
    wafer_slot="w42",
    bands=["SAT_f030",],
    sample_rate=10.0 * u.Hz,
    detset_key="pixel",
    temp_dir=None,
    el_nod=False,
    el_nods=[-1 * u.degree, 1 * u.degree],
    thin_fp=4,
    cal_schedule=False,
):
    """Create a data object with a simple ground sim.

    Use the specified MPI communicator to attempt to create 2 process groups.  Create
    a fake telescope and run the ground sim to make some observations for each
    group.  This is useful for testing many operators that need some pre-existing
    observations with boresight pointing.

    Args:
        mpicomm (MPI.Comm): the MPI communicator (or None).
        sample_rate (Quantity): the sample rate.

    Returns:
        toast.Data: the distributed data with named observations.

    """
    if not toast_available:
        raise RuntimeError("TOAST is not importable, cannot simulate test data")

    # Create the communicator
    toastcomm = create_comm(mpicomm)

    hwp_rpm = 120.0
    hwp_name = defaults.hwp_angle
    if telescope_name == "LAT":
        hwp_name = None
        hwp_rpm = None

    # Simulated telescope
    telescope = sotoast.simulated_telescope(
        hw=None,
        hwfile=None,
        telescope_name=telescope_name,
        sample_rate=sample_rate,
        bands=bands,
        wafer_slots=wafer_slot,
        thinfp=thin_fp,
        comm=mpicomm,
    )

    data = toast.Data(toastcomm)

    # Create a schedule.

    if cal_schedule:
        schedule = calibration_schedule(telescope)
    else:
        schedule = observing_schedule(telescope, mpicomm=toastcomm.comm_world)

    # Add boresight rotation to one scan

    schedule.scans[0].boresight_angle = 30 * u.deg

    sim_ground = toast.ops.SimGround(
        name="sim_ground",
        telescope=telescope,
        session_split_key="tele_wf_band",
        schedule=schedule,
        hwp_angle=hwp_name,
        hwp_rpm=hwp_rpm,
        weather="atacama",
        median_weather=True,
        detset_key=detset_key,
        elnod_start=el_nod,
        elnods=el_nods,
        scan_accel_az=3 * u.degree / u.second ** 2,
        use_ephem=False,
        use_qpoint=True,
    )
    sim_ground.apply(data)

    # corotator = so_ops.CoRotator(name="corotate_lat")
    # corotator.apply(data)

    return data


def simulation_test_multitube(
    mpicomm,
    telescope_name="LAT",
    tubes=None,
    sample_rate=1.0 * u.Hz,
    detset_key="pixel",
    temp_dir=None,
    el_nod=False,
    el_nods=[-1 * u.degree, 1 * u.degree],
    thin_fp=30,
):
    """Create a data object with a simple ground sim.

    This function generates data for a complete telescope at low sample rate, in order
    to test operations that need this.

    Args:
        mpicomm (MPI.Comm): the MPI communicator (or None).
        telescope_name (str):  the telescope to simulate.
        sample_rate (Quantity): the sample rate.

    Returns:
        toast.Data: the distributed data with named observations.

    """
    if not toast_available:
        raise RuntimeError("TOAST is not importable, cannot simulate test data")

    # Create the communicator with a single group
    toastcomm = toast.Comm(world=mpicomm)

    # Simulated telescope
    telescope = sotoast.simulated_telescope(
        hw=None,
        hwfile=None,
        telescope_name=telescope_name,
        tube_slots=tubes,
        sample_rate=sample_rate,
        thinfp=thin_fp,
        comm=mpicomm,
    )

    data = toast.Data(toastcomm)

    # Create a schedule.

    schedule = observing_schedule(telescope, mpicomm=toastcomm.comm_world)

    hwp_name = None
    if telescope_name != "LAT":
        hwp_name = defaults.hwp_angle

    sim_ground = toast.ops.SimGround(
        name="sim_ground",
        telescope=telescope,
        session_split_key="tele_wf_band",
        schedule=schedule,
        weather="atacama",
        median_weather=True,
        detset_key=detset_key,
        elnod_start=el_nod,
        elnods=el_nods,
        scan_accel_az=3 * u.degree / u.second ** 2,
        hwp_angle=hwp_name,
    )
    sim_ground.apply(data)

    return data

# Support functions for a simple obs-like AxisManager...

def quick_dets_axis(n_det):
    return core.LabelAxis('dets', ['det%04i' % i for i in range(n_det)])

def quick_focal_plane(dets, scale=0.5):
    n_det = dets.count
    nrow = int(np.ceil((n_det/2)**2))
    entries = []
    gamma = 0.
    for r in range(nrow):
        y = r * scale
        for c in range(nrow):
            x = c * scale
            entries.extend([(x, y, gamma),
                            (x, y, gamma + 90)])
            gamma += 15.
    entries = entries[(n_det - len(entries)) // 2:]
    entries = entries[:n_det]
    fp = core.AxisManager(dets)
    for k, v in zip(['xi', 'eta', 'gamma'], np.transpose(entries)):
        v = (v - v.mean()) * coords.DEG
        fp.wrap(k, v, [(0, 'dets')])
    return fp

def quick_scan(tod):
    az_min = 100.
    az_max = 120.
    v_az = 2.
    dt = tod.timestamps - tod.timestamps[0]
    az = az_min + (v_az * dt) % (az_max - az_min)
    bs = core.AxisManager(tod.samps)
    bs.wrap_new('az'  , shape=('samps', ))[:] = az * coords.DEG
    bs.wrap_new('el'  , shape=('samps', ))[:] = 50 * coords.DEG
    bs.wrap_new('roll', shape=('samps', ))[:] = 0. * az
    return bs

def quick_tod(n_det, n_samp):
    dets = quick_dets_axis(n_det)
    tod = core.AxisManager(
        quick_dets_axis(n_det),
        core.OffsetAxis('samps', n_samp))
    tod.wrap('focal_plane', quick_focal_plane(tod.dets))
    DT = .1
    dt = np.arange(n_samp) * DT
    tod.wrap_new('timestamps', shape=('samps', ))[:] = 1800000000. + dt
    f_hwp = 2.
    n_hwp = np.ceil(DT * n_samp * f_hwp)
    f_hwp = n_hwp / (DT * n_samp)
    v = DT * n_samp
    tod.wrap_new('hwp_angle',  shape=('samps', ))[:] = (dt * f_hwp + 1.32) % (np.pi*2)
    tod.wrap('boresight', quick_scan(tod))
    tod.wrap_new('signal', shape=('dets', 'samps'), dtype='float32')
    return tod



import numpy as np
import logging
logger = logging.getLogger(__name__)

import so3g
from so3g.proj import coords, quat
import sotodlib.coords.planets as planets

def get_sso(aman, sso, nstep=100):
    """
    Function for getting xi, eta position of given sso.

    Parameters
    ----------
    aman : AxisManager
        Input axis manager.
    sso : str
        Name of input sso.
    nstep : int
        Number of steps to downsample the TOD.

    Returns
    -------
    xi : array
        Array of xi positions.
    eta : array
        Array of eta positions.
    """

    az = aman.boresight.az
    el = aman.boresight.el
    ctime = aman.timestamps
    csl = so3g.proj.CelestialSightLine.az_el(ctime[::nstep], az[::nstep], el[::nstep],
                                             roll=aman.boresight.roll[0], weather="typical")
    q_bore = csl.Q

    ras, decs = [], []
    point_sources = {'tauA': (83.6272579, 22.02159891),
                     'rcw38': (134.7805107, -47.50911231),
                     'iras16183-4958': (245.5154, -50.09168292),
                     'iras19078+0901': (287.5575891, 9.107188994),
                     'rcw122': (260.0339538, -38.95673421),
                     'cenA': (201.3625336, -43.00797508),
                     '3c279': (194.0409868, -5.79174024),
                     '3c273': (187.2775626, 2.053532671),
                     'G025.36-00.14': (279.5264042, -6.793169326),
                     'QSO_J2253+1608': (343.4952422, 16.14301323),
                     'galactic_center': (-93.5833, -29.0078)}
    for d1_unix in ctime[::nstep]:
        if sso in point_sources:
            ra = point_sources[sso][0]
            dec = point_sources[sso][1]
            planet = planets.SlowSource(d1_unix*1., float(ra) * coords.DEG,
                                        float(dec) * coords.DEG)
        else:
            try:
                planet = planets.SlowSource.for_named_source(sso, d1_unix*1.)
            except ValueError:
                break
        ra0, dec0 = planet.pos(d1_unix)
        ras.append(ra0)
        decs.append(dec0)

    planet_q = quat.rotation_lonlat(np.array(ras), np.array(decs))
    if len(q_bore) != len(planet_q):
        raise ValueError("Q vectors have inconsistent lengths")
    q_total = ~q_bore * planet_q
    xi_p, eta_p, _ = quat.decompose_xieta(np.array(q_total))

    return xi_p, eta_p


# Copyright (c) 2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Check functionality of the wiregrid simulation

"""

import unittest
import numpy as np

import astropy.units as u

try:
    # Import sotodlib toast module first, which sets global toast defaults
    import sotodlib.toast as sotoast
    import sotodlib.toast.ops as so_ops
    import toast
    from toast.observation import default_values as defaults
    toast_available = True
except ImportError as e:
    toast_available = False

from ._helpers import calibration_schedule, close_data_and_comm, simulation_test_data


class SimReadoutTest(unittest.TestCase):
    def test_sim_readout(self):
        if not toast_available:
            print("toast cannot be imported- skipping unit tests", flush=True)
            return

        comm, procs, rank = toast.get_world()
        data = simulation_test_data(
            comm,
            telescope_name="LAT",
            wafer_slot="w17",
            bands="LAT_f090",
            sample_rate=10.0 * u.Hz,
            thin_fp=64,
            cal_schedule=False,
        )

        # Simple test just confirms that the operator functions

        so_ops.SimReadout(
            name="sim_readout",
            simulate_glitches=True,
            simulate_jumps=True,
            misidentify_bolometers=True,
            simulate_yield=True,
        ).apply(data)

        close_data_and_comm(data)

if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2018-2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Plot a hardware model.
"""

import argparse

from ..core import Hardware

from ..vis_hardware import summary_text


def main():
    parser = argparse.ArgumentParser(
        description="This program reads a hardware model and prints some\
            summary text to the terminal.",
        usage="so_hardware_info [hardware file, [hardware_file]] ...")

    parser.add_argument(
        "hardware", type=str, nargs="+",
        help="Input hardware file"
    )

    args = parser.parse_args()

    for file in args.hardware:
        print("Loading hardware file {}...".format(file), flush=True)
        hw = Hardware(file)
        summary_text(hw)

    return



# Copyright (c) 2018-2021 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Simons Observatory TOD Simulation and Processing.

This package contains tools for simulating and analyzing data from the
Simons Observatory.  It uses multiple external dependencies (some optional).
See documentation of the various sub-modules for more details.

Contents:

    * core:  Container classes for data, metadata, etc.
    * io:  Code for reading / writing data.
    * utils:  General utility functions and data helper functions.
    * scripts:  Commandline entry points.
    * Top level packages for actual data processing.

"""

from . import _version
__version__ = _version.get_versions()['version']


import logging
logger = logging.getLogger(__name__)
_log_fmt = '%(levelname)s: %(name)s: %(message)s'
_ch = logging.StreamHandler()
_ch.setLevel(logging.INFO)
_ch.setFormatter(logging.Formatter(_log_fmt))
logger.addHandler(_ch)



# Copyright (c) 2018-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import os
import pickle

import ephem
import h5py
import healpy as hp
import numpy as np
import traitlets
from astropy import constants
from astropy import units as u
from scipy.constants import au as AU
from scipy.interpolate import RectBivariateSpline
from toast import qarray as qa
from toast.data import Data
from toast.observation import default_values as defaults
from toast.ops.operator import Operator
from toast.timing import function_timer, Timer
from toast.traits import (Bool, Float, Instance, Int, Quantity, Unicode,
                          trait_docs)
from toast.utils import Environment, Logger, unit_conversion

XAXIS, YAXIS, ZAXIS = np.eye(3)


@trait_docs
class SimWireGrid(Operator):
    """Simulate wiregrid calibration signal"""

    API = Int(0, help="Internal interface version for this operator")

    times = Unicode(defaults.times, help="Observation shared key for timestamps")

    shared_flags = Unicode(
        defaults.shared_flags,
        allow_none=True,
        help="Observation shared key for telescope flags to use",
    )

    shared_flag_mask_wiregrid = Int(
        defaults.shared_mask_invalid, help="Bit mask value for optional shared flagging"
    )

    shared_flag_mask_unstable = Int(
        255, help="Bit mask value applied during wire grid rotation"
    )

    wiregrid_angle = Unicode(
        "wiregrid_angle", help="Observation shared key for wiregrid angle"
    )

    wiregrid_angle_start = Quantity(
        0 * u.degree, help="Starting orientation of the wiregrid"
    )

    wiregrid_step_size = Quantity(22.5 * u.degree, help="Step between orientations")

    wiregrid_step_length = Quantity(10 * u.s, help="Time between orientations")

    wiregrid_angular_speed = Quantity(180 * u.degree / u.second, help="Rotation speed")

    wiregrid_angular_acceleration = Quantity(
        180 * u.degree / u.second**2, help="Rotation acceleration"
    )

    det_data = Unicode(
        defaults.det_data,
        help="Observation detdata key for simulated signal",
    )

    detector_pointing = Instance(
        klass=Operator,
        allow_none=True,
        help="Operator that translates boresight Az/El pointing into detector frame",
    )

    detector_weights = Instance(
        klass=Operator,
        allow_none=True,
        help="Operator that translates boresight Az/El pointing into detector weights",
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @function_timer
    def _get_wiregrid_angle(self, obs):
        """Simulate the motion of the wiregrid and raise appropriate telescope flags"""
        times = obs.shared[self.times].data

        comm = obs.comm_row
        if comm is None or comm.size == 1:
            times_tot = times
        else:
            times_tot = np.hstack(comm.allgather(times))

        t_accel = self.wiregrid_angular_speed / self.wiregrid_angular_acceleration
        dist_accel = 0.5 * self.wiregrid_angular_acceleration * t_accel**2
        if 2 * dist_accel < self.wiregrid_step_size:
            # wiregrid reaches coasting rate between acceleration and deceleration
            dist_coast = self.wiregrid_step_size - 2 * dist_accel
            t_coast = dist_coast / self.wiregrid_angular_speed
        else:
            dist_accel = self.wiregrid_step_size / 2
            t_accel = np.sqrt(2 * dist_accel / self.wiregrid_angular_acceleration)
            t_coast = 0 * u.second
        dist_coast = t_coast * self.wiregrid_angular_speed

        t_static = self.wiregrid_step_length.to_value(u.second)
        t_accel = t_accel.to_value(u.second)
        t_coast = t_coast.to_value(u.second)

        # Simulate the wiregrid angle for the entire observation,
        # one step at a time
        angles = np.zeros(times_tot.size)
        flags = (
            np.zeros(times_tot.size, dtype=np.uint8) + self.shared_flag_mask_wiregrid
        )
        t = times_tot[0]
        istep = 0
        sample = 0
        angle = self.wiregrid_angle_start
        while t < times_tot[-1]:
            # Stationary
            ind = np.logical_and(times_tot >= t, times_tot < t + t_static)
            angles[ind] = angle.to_value(u.radian)
            t += t_static

            # Accelerating
            ind = np.logical_and(times_tot >= t, times_tot < t + t_accel)
            t_ind = (times_tot[ind] - t) * u.second
            angles[ind] = (
                angle + 0.5 * self.wiregrid_angular_acceleration * t_ind**2
            ).to_value(u.radian)
            flags[ind] |= self.shared_flag_mask_unstable
            t += t_accel
            angle += dist_accel
            if t_coast > 0:
                rate = self.wiregrid_angular_speed

                # Coasting
                ind = np.logical_and(times_tot >= t, times_tot < t + t_coast)
                t_ind = (times_tot[ind] - t) * u.second
                angles[ind] = (angle + self.wiregrid_angular_speed * t_ind).to_value(
                    u.radian
                )
                flags[ind] |= self.shared_flag_mask_unstable
                t += t_coast
                angle += dist_coast
            else:
                rate = t_accel * u.second * self.wiregrid_angular_acceleration

            # Decelerating
            ind = np.logical_and(times_tot >= t, times_tot < t + t_accel)
            t_ind = (times_tot[ind] - t) * u.second
            angles[ind] = (
                angle
                + rate * t_ind
                - 0.5 * self.wiregrid_angular_acceleration * t_ind**2
            ).to_value(u.radian)
            flags[ind] |= self.shared_flag_mask_unstable
            t += t_accel
            angle += dist_accel

            # Start the next step
            istep += 1

        # Extract the local piece of the simulated angle
        n_sample = obs.n_local_samples
        istart = obs.local_index_offset
        ind = slice(istart, istart + n_sample)
        angles = angles[ind]
        flags = flags[ind]

        # Deposit in the Observation
        obs.shared.create_column(
            self.wiregrid_angle, shape=(n_sample,), dtype=np.float64
        )
        obs.shared[self.wiregrid_angle].set(angles, offset=(0,), fromrank=0)
        obs.shared[self.shared_flags].data |= flags

        return

    def get_wiregrid_signal(self, obs, band):
        wiregrid_angle = obs.shared[self.wiregrid_angle].data

        # FIXME: figure out the actual amplitude of the signal in K_CMB

        amplitude = 1

        I = amplitude * np.ones_like(wiregrid_angle)
        Q = amplitude * np.cos(2 * wiregrid_angle)
        U = amplitude * np.sin(2 * wiregrid_angle)

        return I, Q, U

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()

        for trait in "detector_pointing", "detector_weights", "shared_flags":
            value = getattr(self, trait)
            if value is None:
                raise RuntimeError(f"You must set `{trait}` before running SimWireGrid")

        for obs in data.obs:
            if "WIREGRID" not in obs.name.upper():
                log.debug_rank(
                    f"SimWireGrid: {obs.name} is not a wiregrid calibration observation.",
                    comm=obs.comm.comm_group,
                )
                continue
            self._get_wiregrid_angle(obs)

            dets = obs.select_local_detectors(detectors)
            obs.detdata.ensure(self.det_data, detectors=dets, create_units=u.K)
            det_units = obs.detdata[self.det_data].units
            scale = unit_conversion(u.K, det_units)

            focalplane = obs.telescope.focalplane
            for det in dets:
                signal = obs.detdata[self.det_data][det]

                # Compute detector quaternions and Stokes weights
                obs_data = Data(comm=data.comm)
                obs_data._internal = data._internal
                obs_data.obs = [obs]
                self.detector_pointing.apply(obs_data, detectors=[det])
                self.detector_weights.apply(obs_data, detectors=[det])
                obs_data.obs.clear()
                del obs_data

                band = focalplane[det]["band"]

                I, Q, U = self.get_wiregrid_signal(obs, band)

                weights = obs.detdata[self.detector_weights.weights][det]
                weight_mode = self.detector_weights.mode
                if "I" in weight_mode:
                    ind = weight_mode.index("I")
                    weights_I = weights[:, ind].copy()
                else:
                    weights_I = 0
                if "Q" in weight_mode:
                    ind = weight_mode.index("Q")
                    weights_Q = weights[:, ind].copy()
                else:
                    weights_Q = 0
                if "U" in weight_mode:
                    ind = weight_mode.index("U")
                    weights_U = weights[:, ind].copy()
                else:
                    weights_U = 0

                signal += scale * (I * weights_I + Q * weights_Q + U * weights_U)

        return

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = self.detector_pointing.requires()
        req["shared"].append(self.times)
        req["detdata"].append(self.weights)
        return req

    def _provides(self):
        prov = {
            "meta": list(),
            "shared": [self.wiregrid_angle],
            "detdata": [self.det_data],
        }
        return prov

    def _accelerators(self):
        return list()
